# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function

from astropy.utils.exceptions import AstropyUserWarning

__all__ = [
    'FITSRegionParserWarning',
    'FITSRegionParserError',
]


class FITSRegionParserWarning(AstropyUserWarning):
    """
    A generic warning class for FITS region parsing
    """


class FITSRegionParserError(ValueError):
    """
    A generic error class for FITS region parsing
    """


language_spec = {'CIRCLE': ['X0', 'Y0', 'R0'],
                 'POINT': ['X0', 'Y0'],
                 'BOX': ['X0', 'Y0', 'R0', 'R1', 'ROTANG0'],
                 'ANNULUS': ['X0', 'Y0', 'R0', 'R1'],
                 'ELLIPSE': ['X0', 'Y0', 'R0', 'R1', 'ROTANG0'],
                 'ELLIPTANNULUS': ['X0', 'Y0', 'R0', 'R1', 'R2', 'R3', 'ROTANG0'],
                 'ROTBOX': ['X0', 'Y0', 'R0', 'R1', 'ROTANG0'],
                 'RECTANGLE': ['X0', 'X1', 'Y0', 'Y1'],
                 'ROTRECTANGLE': ['X0', 'X1', 'Y0', 'Y1', 'ROTANG0'],
                 'POLYGON': ['X', 'Y'],
                 'PIE': ['X0', 'Y0', 'ROTANG0', 'ROTANG1']
                 }
language_spec['SECTOR'] = language_spec['PIE']



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Region shapes.
"""
from .circle import *
from .ellipse import *
from .point import *
from .polygon import *
from .rectangle import *
from .annulus import *
from .line import *
from .text import *



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This is an in-development package for region handling based on Astropy.

The goal is to merge the functionality from pyregion and photutils apertures
and then after some time propose this package for inclusion in the Astropy core.

* Code : https://github.com/astropy/regions
* Docs : http://astropy-regions.readthedocs.io/en/latest/
"""

# Affiliated packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._astropy_init import *
# ----------------------------------------------------------------------------

# For egg_info test builds to pass, put package imports here.
if not _ASTROPY_SETUP_:
    from ._utils.examples import *
    from .core import *
    from .shapes import *
    from .io import *



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from numpy.testing import assert_allclose
import pytest

from ..bounding_box import BoundingBox

try:
    import matplotlib
    HAS_MATPLOTLIB = True
except:
    HAS_MATPLOTLIB = False


def test_bounding_box_init():
    bbox = BoundingBox(1, 10, 2, 20)

    assert bbox.ixmin == 1
    assert bbox.ixmax == 10
    assert bbox.iymin == 2
    assert bbox.iymax == 20


def test_bounding_box_init_minmax():
    with pytest.raises(ValueError):
        BoundingBox(100, 1, 1, 100)
    with pytest.raises(ValueError):
        BoundingBox(1, 100, 100, 1)


def test_bounding_box_inputs():
    with pytest.raises(TypeError):
        BoundingBox([1], [10], [2], [9])
    with pytest.raises(TypeError):
        BoundingBox([1, 2], 10, 2, 9)
    with pytest.raises(TypeError):
        BoundingBox(1.0, 10.0, 2.0, 9.0)
    with pytest.raises(TypeError):
        BoundingBox(1.3, 10, 2, 9)
    with pytest.raises(TypeError):
        BoundingBox(1, 10.3, 2, 9)
    with pytest.raises(TypeError):
        BoundingBox(1, 10, 2.3, 9)
    with pytest.raises(TypeError):
        BoundingBox(1, 10, 2, 9.3)


def test_bounding_box_from_float():
    # This is the example from the method docstring
    bbox = BoundingBox.from_float(xmin=1.0, xmax=10.0, ymin=2.0, ymax=20.0)
    assert bbox == BoundingBox(ixmin=1, ixmax=11, iymin=2, iymax=21)

    bbox = BoundingBox.from_float(xmin=1.4, xmax=10.4, ymin=1.6, ymax=10.6)
    assert bbox == BoundingBox(ixmin=1, ixmax=11, iymin=2, iymax=12)


def test_bounding_box_eq():
    bbox = BoundingBox(1, 10, 2, 20)
    assert bbox == bbox

    assert bbox != BoundingBox(9, 10, 2, 20)
    assert bbox != BoundingBox(1, 99, 2, 20)
    assert bbox != BoundingBox(1, 10, 9, 20)
    assert bbox != BoundingBox(1, 10, 2, 99)


def test_bounding_box_repr():
    bbox = BoundingBox(1, 10, 2, 20)

    assert repr(bbox) == 'BoundingBox(ixmin=1, ixmax=10, iymin=2, iymax=20)'
    assert eval(repr(bbox)) == bbox


def test_bounding_box_shape():
    bbox = BoundingBox(1, 10, 2, 20)

    assert bbox.shape == (18, 9)


def test_bounding_box_slices():
    bbox = BoundingBox(1, 10, 2, 20)

    assert bbox.slices == (slice(2, 20), slice(1, 10))


def test_bounding_box_extent():
    bbox = BoundingBox(1, 10, 2, 20)

    assert_allclose(bbox.extent, (0.5, 9.5, 1.5, 19.5))


@pytest.mark.skipif('not HAS_MATPLOTLIB')
def test_bounding_box_as_artist():
    bbox = BoundingBox(1, 10, 2, 20)

    patch = bbox.as_artist()
    assert_allclose(patch.get_xy(), (0.5, 1.5))
    assert_allclose(patch.get_width(), 9)
    assert_allclose(patch.get_height(), 18)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function

from warnings import warn

from astropy import units as u
import numpy as np
from astropy.table import Table
from astropy.io import fits
from astropy.wcs import WCS

from .core import FITSRegionParserError, FITSRegionParserWarning, language_spec
from ..core import Shape, ShapeList, reg_mapping

__all__ = ['FITSRegionParser', 'read_fits_region', 'FITSRegionRowParser']


class FITSRegionParser(object):
    """
    Parses a FITS Region table.

    Parameters
    ----------
    table: `~astropy.table.Table` object
        A fits region table
    errors : ``warn``, ``ignore``, ``strict``
        The error handling scheme to use for handling parsing errors.
        The default is 'strict', which will raise a `FITSRegionParserError`.
        `warn`` will raise a `FITSRegionParserWarning`, and ``ignore`` will
        do nothing (i.e., be silent).

    Examples
    --------

    >>> from regions import FITSRegionParser
    >>> from astropy.table import Table
    >>> from astropy.utils.data import get_pkg_data_filename
    >>> filename = get_pkg_data_filename('data/regions.fits',
    ...                                  package='regions.io.fits.tests')
    >>> table = Table.read(filename)
    >>> print(table)
        X         Y      SHAPE          R [2]            ROTANG  COMPONENT
      pix       pix                     pix              deg
    --------- --------- ------- ---------------------- --------- ---------
       2896.5    5056.5  CIRCLE        381.9716 .. 0.0       0.0         1
    5282.0541 4854.5699     BOX 1303.4597 .. 655.09466 28.395178         2
       2944.5    3472.5 ELLIPSE         288.0 .. 512.0  337.4048         3
        341.0     345.0  ROTBOX           56.0 .. 78.0      65.0         4
        341.0     345.0 ANNULUS           56.0 .. 78.0       0.0         5
        341.0     345.0   POINT             0.0 .. 0.0       0.0         6
        341.0     345.0                     0.0 .. 0.0       0.0         7

    >>> parser = FITSRegionParser(table)
    >>> shapes = parser.shapes
    >>> regions = shapes.to_regions()

    """

    valid_columns = ['X', 'Y', 'SHAPE', 'COMPONENT', 'R', 'ROTANG']

    def __init__(self, table, errors='strict'):

        if errors not in ('strict', 'ignore', 'warn'):
            msg = "``errors`` must be one of strict, ignore, or warn; is {}"
            raise ValueError(msg.format(errors))
        if not isinstance(table, Table):
            raise TypeError("The table should be an astropy table object")
        self.table = table
        self.errors = errors
        self.unit = {}
        self._shapes = {}

        self.parse_table()

    @property
    def shapes(self):
        shape_list = ShapeList()
        components = list(self._shapes.keys())
        components.sort()
        for component in components:
            shape_list += ShapeList(self._shapes[component])
        return shape_list

    def _raise_error(self, msg):
        if self.errors == 'warn':
            warn(msg, FITSRegionParserWarning)
        elif self.errors == 'strict':
            raise FITSRegionParserError(msg)

    def parse_table(self):

        for col in self.table.colnames:
            if col not in self.valid_columns:
                self._raise_error("This table has an invalid column name: '{}'"
                                  .format(col))
            else:
                self.unit[col] = self.table[col].unit

        if 'COMPONENT' not in self.table.colnames:
            self.table['COMPONENT'] = np.ones([1, len(self.table)])

        for row in self.table:
            reg = FITSRegionRowParser(row, self.unit, self.errors)
            component, shape = reg.parse()
            if component in self.shapes:
                self._shapes[component].append(shape)
            else:
                self._shapes[component] = [shape]


class FITSRegionRowParser():
    """
    Parses a single row of the FITS region table

    Parameters
    ----------
    row: `~astropy.table.row.Row` object
        Single row of the region table that is to be parsed.
    unit: `dict`
        Units of each column in the row.
    errors : ``warn``, ``ignore``, ``strict``
        The error handling scheme to use for handling parsing errors.
        The default is 'strict', which will raise a ``FITSRegionParserError``.
        `warn`` will raise a ``FITSRegionParserWarning``, and ``ignore`` will
        do nothing (i.e., be silent).

    """
    def __init__(self, row, unit, errors='strict'):
        self.errors = errors
        self.unit = unit
        self.row = row

        region_type = self._get_col_value('SHAPE0', 'POINT')[0]

        # Default region is POINT
        if region_type == '':
            region_type = 'POINT'

        if region_type[0] == '!':
            self.include = False
            region_type = region_type[1:]
        else:
            self.include = True

        region_type = region_type.strip().upper()

        if region_type in language_spec:
            self.region_type = region_type
        else:
            self._raise_error("'{0}' is not a valid FITS Region type"
                              .format(region_type))

        self.component = str(self.row['COMPONENT'])

    def _get_col_value(self, colname, default=None):
        index = None

        if colname[-1].isdigit():
            index = int(colname[-1])
            colname = colname[: -1]

        try:
            val, unit = self.row[colname], self.unit[colname]

            if np.isscalar(val):
                val = np.array(val).reshape(1, )
            if index is not None:
                if index < len(val) and val[index] != 0:
                    return val[index], unit
                else:
                    raise ValueError("The column: {0} must have more than {1} value for the "
                                     "region {2}".format(colname, index,
                                                        self.region_type))
            else:
                return val, unit
        except KeyError:
            if default is None:
                self._raise_error("The column: '{0}' is missing in the table"
                                  .format(colname))
            else:
                return default

    def _raise_error(self, msg):
        if self.errors == 'warn':
            warn(msg, FITSRegionParserWarning)
        elif self.errors == 'strict':
            raise FITSRegionParserError(msg)

    def parse(self):

        coords = []

        for x in language_spec[self.region_type]:
            y, unit = self._get_col_value(x)
            coords.append(self._parse_value(y, unit))

        meta = {'tag': self.component}

        if self.region_type == 'POLYGON':
            coords_new = []
            for x, y in zip(coords[0], coords[1]):
                coords_new += [x, y]
            coords = coords_new

        region_type = self.region_type.lower()
        if region_type in reg_mapping['FITS_REGION']:
            region_type = reg_mapping['FITS_REGION'][region_type]
        else:
            self._raise_error("'{0}' is currently not supported in "
                              "regions".format(self.region_type))

        return self.component, Shape('physical', region_type,
                                     coords, meta, False, False)

    def _parse_value(self, val, unit):

        units = dict(pix=u.dimensionless_unscaled,
                     deg=u.deg,
                     rad=u.rad,
                     )

        if unit is not None:
            return val * units.get(str(unit), unit)
        else:
            self._raise_error("The unit: {} is invalid".format(unit))


def read_fits_region(filename, errors='strict'):
    """
    Reads a FITS region file and scans for any fits regions table and
    converts them into `Region` objects.

    Parameters
    ----------
    filename : str
        The file path
    errors : ``warn``, ``ignore``, ``strict``
      The error handling scheme to use for handling parsing errors.
      The default is 'strict', which will raise a `FITSRegionParserError`.
      ``warn`` will raise a `FITSRegionParserWarning`, and ``ignore`` will do nothing
      (i.e., be silent).

    Returns
    -------
    regions : list
        Python list of `regions.Region` objects.

    Examples
    --------

    >>> from astropy.utils.data import get_pkg_data_filename
    >>> from regions import read_fits_region
    >>> file_read = get_pkg_data_filename('data/region.fits',
    ...                                   package='regions.io.fits.tests')
    >>> regions = read_fits_region(file_read)

    """
    regions = []

    hdul = fits.open(filename)

    for hdu in hdul:
        if hdu.name == 'REGION':
            table = Table.read(hdu)
            wcs = WCS(hdu.header, keysel=['image', 'binary', 'pixel'])
            regions_list = FITSRegionParser(table, errors).shapes.to_regions()
            for reg in regions_list:
                regions.append(reg.to_sky(wcs))

    return regions



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This file sets up detailed tests for computing masks with reference images.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import itertools
import pytest

from astropy import units as u

from ...core import PixCoord
from ...shapes.circle import CirclePixelRegion
from ...shapes.ellipse import EllipsePixelRegion
from ...shapes.rectangle import RectanglePixelRegion
from ...shapes.polygon import PolygonPixelRegion

REGIONS = [
    CirclePixelRegion(PixCoord(3.981987, 4.131378), radius=3.3411),
    EllipsePixelRegion(PixCoord(5.981987, 4.131378), width=10.4466, height=6.6822, angle=12 * u.deg),
    RectanglePixelRegion(PixCoord(5.981987, 4.131378), width=7.2233, height=4.3411, angle=12 * u.deg),
    PolygonPixelRegion(PixCoord([-2.334, 3.631, 1.122, -1.341], [-3.121, -2.118, 2.987, 1.759])),
]

MODES = [
    {'mode': 'center'},
    {'mode': 'exact'},
    {'mode': 'subpixels', 'subpixels': 1},
    {'mode': 'subpixels', 'subpixels': 5},
    {'mode': 'subpixels', 'subpixels': 18},
]


def label(value):
    if isinstance(value, CirclePixelRegion):
        return 'circ'
    elif isinstance(value, EllipsePixelRegion):
        return 'elli'
    elif isinstance(value, RectanglePixelRegion):
        return 'rect'
    elif isinstance(value, PolygonPixelRegion):
        return 'poly'
    else:
        return '-'.join('{0}_{1}'.format(key, value) for key, value in sorted(value.items()))

# There's a bug in numpy: `numpy.savetxt` doesn't accept unicode
# on Python 2. Bytes works on Python 2 and 3, so we're using that here.
# https://github.com/numpy/numpy/pull/4053#issuecomment-263808221
@pytest.mark.array_compare(
    fmt='text',
    write_kwargs={'fmt': b'%12.8e'},
)
@pytest.mark.parametrize(
    ('region', 'mode'),
    itertools.product(REGIONS, MODES),
    ids=label,
)
def test_to_mask(region, mode):
    try:
        mask = region.to_mask(**mode)
    except NotImplementedError:
        pytest.xfail()
    return mask.data.astype(float)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
from numpy.testing import assert_allclose
import pytest

from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.utils.data import get_pkg_data_filename
from astropy.io import fits
from astropy.wcs import WCS

from ...core import PixCoord
from ...tests.helpers import make_simple_wcs
from ..point import PointPixelRegion, PointSkyRegion
from .utils import ASTROPY_LT_13, HAS_MATPLOTLIB  # noqa
from .test_common import BaseTestPixelRegion, BaseTestSkyRegion


@pytest.fixture(scope='session')
def wcs():
    filename = get_pkg_data_filename('data/example_header.fits')
    header = fits.getheader(filename)
    return WCS(header)


class TestPointPixelRegion(BaseTestPixelRegion):

    reg = PointPixelRegion(PixCoord(3, 4))
    sample_box = [-2, 8, -1, 9]
    inside = []
    outside = [(3.1, 4.2), (5, 4)]
    expected_area = 0
    expected_repr = '<PointPixelRegion(PixCoord(x=3, y=4))>'
    expected_str = 'Region: PointPixelRegion\ncenter: PixCoord(x=3, y=4)'

    def test_pix_sky_roundtrip(self):
        wcs = make_simple_wcs(SkyCoord(2 * u.deg, 3 * u.deg), 0.1 * u.deg, 20)
        reg_new = self.reg.to_sky(wcs).to_pixel(wcs)
        assert_allclose(reg_new.center.x, self.reg.center.x)
        assert_allclose(reg_new.center.y, self.reg.center.y)

    @pytest.mark.skipif('not HAS_MATPLOTLIB')
    def test_as_artist(self):
        artist = self.reg.as_artist()

        assert artist.get_data() == ([3], [4])

        artist = self.reg.as_artist(origin=(1, 1))

        assert artist.get_data() == ([2], [3])


class TestPointSkyRegion(BaseTestSkyRegion):

    reg = PointSkyRegion(SkyCoord(3, 4, unit='deg'))

    if ASTROPY_LT_13:
        expected_repr = ('<PointSkyRegion(<SkyCoord (ICRS): (ra, dec) in deg\n'
                         '    (3.0, 4.0)>)>')
        expected_str = ('Region: PointSkyRegion\ncenter: <SkyCoord (ICRS): '
                        '(ra, dec) in deg\n    (3.0, 4.0)>')
    else:
        expected_repr = ('<PointSkyRegion(<SkyCoord (ICRS): (ra, dec) in deg\n'
                         '    ( 3.,  4.)>)>')
        expected_str = ('Region: PointSkyRegion\ncenter: <SkyCoord (ICRS): '
                        '(ra, dec) in deg\n    ( 3.,  4.)>')

    def test_contains(self, wcs):
        position = SkyCoord([1, 2] * u.deg, [3, 4] * u.deg)
        # points do not contain things
        assert all(self.reg.contains(position, wcs) == np.array([False,False], dtype='bool'))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import numpy as np
from astropy.io.fits.util import _is_int


__all__ = ['BoundingBox']


class BoundingBox(object):
    """
    A rectangular bounding box in integer (not float) pixel indices.

    Parameters
    ----------
    ixmin, ixmax, iymin, iymax : int
        The bounding box pixel indices.  Note that the upper values
        (``iymax`` and ``ixmax``) are exclusive as for normal slices in
        Python.  The lower values (``ixmin`` and ``iymin``) must not be
        greater than the respective upper values (``ixmax`` and
        ``iymax``).

    Examples
    --------
    >>> from regions import BoundingBox

    >>> # constructing a BoundingBox like this is cryptic:
    >>> bbox = BoundingBox(1, 10, 2, 20)

    >>> # it's better to use keyword arguments for readability:
    >>> bbox = BoundingBox(ixmin=1, ixmax=10, iymin=2, iymax=20)
    >>> bbox  # nice repr, useful for interactive work
    BoundingBox(ixmin=1, ixmax=10, iymin=2, iymax=20)

    >>> # sometimes it's useful to check if two bounding boxes are the same
    >>> bbox == BoundingBox(ixmin=1, ixmax=10, iymin=2, iymax=20)
    True
    >>> bbox == BoundingBox(ixmin=7, ixmax=10, iymin=2, iymax=20)
    False

    >>> # "shape" and "slices" can be useful when working with numpy arrays
    >>> bbox.shape  # numpy order: (y, x)
    (18, 9)
    >>> bbox.slices  # numpy order: (y, x)
    (slice(2, 20, None), slice(1, 10, None))

    >>> # "extent" is useful when plotting the BoundingBox with matplotlib
    >>> bbox.extent  # matplotlib order: (x, y)
    (0.5, 9.5, 1.5, 19.5)
    >>> print(bbox.as_artist())
    Rectangle(0.5,1.5;9x18)
    """

    def __init__(self, ixmin, ixmax, iymin, iymax):
        if not _is_int(ixmin):
            raise TypeError('ixmin must be an integer')
        if not _is_int(ixmax):
            raise TypeError('ixmax must be an integer')
        if not _is_int(iymin):
            raise TypeError('iymin must be an integer')
        if not _is_int(iymax):
            raise TypeError('iymax must be an integer')

        if ixmin > ixmax:
            raise ValueError('ixmin must be <= ixmax')
        if iymin > iymax:
            raise ValueError('iymin must be <= iymax')

        self.ixmin = ixmin
        self.ixmax = ixmax
        self.iymin = iymin
        self.iymax = iymax

    @classmethod
    def from_float(cls, xmin, xmax, ymin, ymax):
        """
        Return the smallest bounding box that fully contains a given
        rectangle defined by float coordinate values.

        Following the pixel index convention, an integer index
        corresponds to the center of a pixel and the pixel edges span
        from (index - 0.5) to (index + 0.5).  For example, the pixel
        edge spans of the following pixels are:

        - pixel 0: from -0.5 to 0.5
        - pixel 1: from 0.5 to 1.5
        - pixel 2: from 1.5 to 2.5

        In addition, because `BoundingBox` upper limits are exclusive
        (by definition), 1 is added to the upper pixel edges.  See
        examples below.

        Parameters
        ----------
        xmin, xmax, ymin, ymax : float
            Float coordinates defining a rectangle.  The lower values
            (``xmin`` and ``ymin``) must not be greater than the
            respective upper values (``xmax`` and ``ymax``).

        Returns
        -------
        bbox : `BoundingBox` object
            The minimal ``BoundingBox`` object fully containing the
            input rectangle coordinates.

        Examples
        --------
        >>> from regions import BoundingBox
        >>> BoundingBox.from_float(xmin=1.0, xmax=10.0, ymin=2.0, ymax=20.0)
        BoundingBox(ixmin=1, ixmax=11, iymin=2, iymax=21)

        >>> BoundingBox.from_float(xmin=1.4, xmax=10.4, ymin=1.6, ymax=10.6)
        BoundingBox(ixmin=1, ixmax=11, iymin=2, iymax=12)
        """

        ixmin = int(np.floor(xmin + 0.5))
        ixmax = int(np.ceil(xmax + 0.5))
        iymin = int(np.floor(ymin + 0.5))
        iymax = int(np.ceil(ymax + 0.5))

        return cls(ixmin, ixmax, iymin, iymax)

    def __eq__(self, other):
        if not isinstance(other, BoundingBox):
            raise TypeError('Can compare BoundingBox only to another '
                            'BoundingBox.')

        return (
            (self.ixmin == other.ixmin) and
            (self.ixmax == other.ixmax) and
            (self.iymin == other.iymin) and
            (self.iymax == other.iymax)
        )

    def __repr__(self):
        data = self.__dict__
        data['name'] = self.__class__.__name__
        fmt = ('{name}(ixmin={ixmin}, ixmax={ixmax}, iymin={iymin}, '
               'iymax={iymax})')
        return fmt.format(**data)

    @property
    def shape(self):
        """
        The ``(ny, nx)`` shape of the bounding box.
        """

        return self.iymax - self.iymin, self.ixmax - self.ixmin

    @property
    def slices(self):
        """
        The bounding box as a tuple of `slice` objects.

        The slice tuple is in numpy axis order (i.e. ``(y, x)``) and
        therefore can be used to slice numpy arrays.
        """

        return (slice(self.iymin, self.iymax), slice(self.ixmin, self.ixmax))

    @property
    def extent(self):
        """
        The extent of the mask, defined as the ``(xmin, xmax, ymin,
        ymax)`` bounding box from the bottom-left corner of the
        lower-left pixel to the upper-right corner of the upper-right
        pixel.

        The upper edges here are the actual pixel positions of the
        edges, i.e. they are not "exclusive" indices used for python
        indexing.  This is useful for plotting the bounding box using
        Matplotlib.
        """

        return (
            self.ixmin - 0.5,
            self.ixmax - 0.5,
            self.iymin - 0.5,
            self.iymax - 0.5,
        )

    def as_artist(self, **kwargs):
        """
        Return a `matplotlib.patches.Rectangle` that represents the
        bounding box.

        Parameters
        ----------
        kwargs : `dict`
            Any keyword arguments accepted by
            `matplotlib.patches.Patch`.

        Returns
        -------
        result : `matplotlib.patches.Rectangle`
            A matplotlib rectangular patch.

        Examples
        --------
        .. plot::
            :include-source:

            import matplotlib.pyplot as plt
            from regions import BoundingBox
            bbox = BoundingBox(2, 7, 3, 8)
            fig = plt.figure()
            ax = fig.add_subplot(1, 1, 1)
            np.random.seed(12345)
            ax.imshow(np.random.random((10, 10)), interpolation='nearest', cmap='viridis')
            ax.add_patch(bbox.as_artist(facecolor='none', edgecolor='white', lw=2.))
        """

        from matplotlib.patches import Rectangle

        return Rectangle(xy=(self.extent[0], self.extent[2]),
                         width=self.shape[1], height=self.shape[0], **kwargs)

    def to_region(self):
        """
        Return a `~regions.RectanglePixelRegion` that
        represents the bounding box.
        """

        from ..shapes import RectanglePixelRegion
        from .pixcoord import PixCoord

        xpos = (self.extent[1] + self.extent[0]) / 2.
        ypos = (self.extent[3] + self.extent[2]) / 2.
        xypos = PixCoord(xpos, ypos)
        h, w = self.shape

        return RectanglePixelRegion(center=xypos, width=w, height=h)

    def plot(self, origin=(0, 0), ax=None, **kwargs):
        """
        Plot the `BoundingBox` on a matplotlib `~matplotlib.axes.Axes`
        instance.

        Parameters
        ----------
        origin : array_like, optional
            The ``(x, y)`` position of the origin of the displayed
            image.
        ax : `matplotlib.axes.Axes` instance, optional
            If `None`, then the current `~matplotlib.axes.Axes` instance
            is used.
        kwargs : `dict`
            Any keyword arguments accepted by `matplotlib.patches.Patch`.

        Returns
        -------
        ax : `~matplotlib.axes.Axes`
            Axes on which the patch is added.
        """

        reg = self.to_region()
        return reg.plot(origin=origin, ax=ax, **kwargs)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# (taken from photutils: should probably migrate into astropy.wcs)
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np
from astropy import units as u
from astropy.coordinates import UnitSphericalRepresentation
from astropy.wcs.utils import skycoord_to_pixel
from ..core.pixcoord import PixCoord

skycoord_to_pixel_mode = 'all'


def skycoord_to_pixel_scale_angle(skycoord, wcs, small_offset=1 * u.arcsec):
    """
    Convert a set of SkyCoord coordinates into pixel coordinates, pixel
    scales, and position angles.

    Parameters
    ----------
    skycoord : `~astropy.coordinates.SkyCoord`
        Sky coordinates
    wcs : `~astropy.wcs.WCS`
        The WCS transformation to use
    small_offset : `~astropy.units.Quantity`
        A small offset to use to compute the angle

    Returns
    -------
    pixcoord : `~regions.PixCoord`
        Pixel coordinates
    scale : float
        The pixel scale at each location, in degrees/pixel
    angle : `~astropy.units.Quantity`
        The position angle of the celestial coordinate system in pixel space.
    """

    # Convert to pixel coordinates
    x, y = skycoord_to_pixel(skycoord, wcs, mode=skycoord_to_pixel_mode)
    pixcoord = PixCoord(x=x, y=y)

    # We take a point directly 'above' (in latitude) the position requested
    # and convert it to pixel coordinates, then we use that to figure out the
    # scale and position angle of the coordinate system at the location of
    # the points.

    # Find the coordinates as a representation object
    r_old = skycoord.represent_as('unitspherical')

    # Add a a small perturbation in the latitude direction (since longitude
    # is more difficult because it is not directly an angle).
    dlat = small_offset
    r_new = UnitSphericalRepresentation(r_old.lon, r_old.lat + dlat)
    coords_offset = skycoord.realize_frame(r_new)

    # Find pixel coordinates of offset coordinates
    x_offset, y_offset = skycoord_to_pixel(coords_offset, wcs,
                                           mode=skycoord_to_pixel_mode)

    # Find vector
    dx = x_offset - x
    dy = y_offset - y

    # Find the length of the vector
    scale = np.hypot(dx, dy) / dlat.to('degree').value

    # Find the position angle
    angle = np.arctan2(dy, dx) * u.radian

    return pixcoord, scale, angle


def assert_angle_or_pixel(name, q):
    """
    Check that ``q`` is either an angular or a pixel `~astropy.units.Quantity`.
    """
    if isinstance(q, u.Quantity):
        if q.unit.physical_type == 'angle' or q.unit is u.pixel:
            pass
        else:
            raise ValueError("{0} should have angular or pixel "
                             "units".format(name))
    else:
        raise TypeError("{0} should be a Quantity instance".format(name))


def assert_angle(name, q):
    """
    Check that ``q`` is an angular `~astropy.units.Quantity`.
    """
    if isinstance(q, u.Quantity):
        if q.unit.physical_type == 'angle':
            pass
        else:
            raise ValueError("{0} should have angular units".format(name))
    else:
        raise TypeError("{0} should be a Quantity instance".format(name))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals

from astropy.io import fits

from ..core import to_shape_list, SkyRegion


__all__ = [
    'write_fits_region',
    'fits_region_objects_to_table',
]


def fits_region_objects_to_table(regions):
    """
    Converts list of regions to FITS region table.

    Parameters
    ----------
    regions : list
        List of `regions.Region` objects

    Returns
    -------
    region_string : `~astropy.table.Table`
        FITS region table

    Examples
    --------

    >>> from regions import CirclePixelRegion, PixCoord
    >>> reg_pixel = CirclePixelRegion(PixCoord(1, 2), 5)
    >>> table = fits_region_objects_to_table([reg_pixel])
    >>> print(table)
    X [1] Y [1] SHAPE    R [4]    ROTANG COMPONENT
    pix   pix            pix      deg
    ----- ----- ------ ---------- ------ ---------
      1.0   2.0 circle 5.0 .. 0.0      0         1

    """
    for reg in regions:
        if isinstance(reg, SkyRegion):
            raise TypeError('Every region must be a pixel region'.format(reg))

    shape_list = to_shape_list(regions, coordinate_system='image')
    return shape_list.to_fits()


def write_fits_region(filename, regions, header=None):
    """
    Converts list of regions to FITS region table and write to a file.

    Parameters
    ----------
    filename: str
        Filename in which the table is to be written. Default is 'new.fits'
    regions: list
        List of `regions.Region` objects
    header: `~astropy.io.fits.header.Header` object
        The FITS header.

    Examples
    --------

    >>> from astropy.utils.data import get_pkg_data_filename
    >>> from astropy.io import fits
    >>> file_sample = get_pkg_data_filename('data/fits_region.fits', package='regions.io.fits.tests')
    >>> from regions import CirclePixelRegion, PixCoord, write_fits_region
    >>> reg_pixel = CirclePixelRegion(PixCoord(1, 2), 5)
    >>> hdul = fits.open(file_sample)
    >>> write_fits_region('region_output.fits', regions=[reg_pixel], header=hdul[1].header)

    """
    output = fits_region_objects_to_table(regions)

    bin_table = fits.BinTableHDU(data=output, header=header)
    bin_table.writeto(filename)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
from numpy.testing import assert_allclose
import pytest

import astropy.units as u
from astropy.tests.helper import assert_quantity_allclose
from astropy.coordinates import SkyCoord
from astropy.utils.data import get_pkg_data_filename
from astropy.io import fits
from astropy.wcs import WCS

from ...core import PixCoord
from ...tests.helpers import make_simple_wcs
from ..ellipse import EllipsePixelRegion, EllipseSkyRegion
from .utils import ASTROPY_LT_13, HAS_MATPLOTLIB  # noqa
from .test_common import BaseTestPixelRegion, BaseTestSkyRegion


@pytest.fixture(scope='session')
def wcs():
    filename = get_pkg_data_filename('data/example_header.fits')
    header = fits.getheader(filename)
    return WCS(header)


class TestEllipsePixelRegion(BaseTestPixelRegion):

    reg = EllipsePixelRegion(center=PixCoord(3, 4), width=4, height=3, angle=5 * u.deg)
    sample_box = [-2, 8, -1, 9]
    inside = [(4.5, 4)]
    outside = [(5, 4)]
    expected_area = 3 * np.pi
    expected_repr = '<EllipsePixelRegion(PixCoord(x=3, y=4), width=4, height=3, angle=5.0 deg)>'
    expected_str = ('Region: EllipsePixelRegion\ncenter: PixCoord(x=3, y=4)\n'
                    'width: 4\nheight: 3\nangle: 5.0 deg')

    def test_pix_sky_roundtrip(self):
        wcs = make_simple_wcs(SkyCoord(2 * u.deg, 3 * u.deg), 0.1 * u.deg, 20)
        reg_new = self.reg.to_sky(wcs).to_pixel(wcs)
        assert_allclose(reg_new.center.x, self.reg.center.x)
        assert_allclose(reg_new.center.y, self.reg.center.y)
        assert_allclose(reg_new.width, self.reg.width)
        assert_allclose(reg_new.height, self.reg.height)
        assert_quantity_allclose(reg_new.angle, self.reg.angle)

    @pytest.mark.skipif('not HAS_MATPLOTLIB')
    def test_as_artist(self):
        patch = self.reg.as_artist()
        assert_allclose(patch.center, (3, 4))
        assert_allclose(patch.width, 4)
        assert_allclose(patch.height, 3)
        assert_allclose(patch.angle, 5)

    def test_region_bbox(self):
        a = 7
        b = 3
        reg = EllipsePixelRegion(PixCoord(50, 50), width=a, height=b,
                                 angle=0. * u.deg)
        assert reg.bounding_box.shape == (b, a)

        reg = EllipsePixelRegion(PixCoord(50.5, 50.5), width=a, height=b,
                                 angle=0. * u.deg)
        assert reg.bounding_box.shape == (b + 1, a + 1)

        reg = EllipsePixelRegion(PixCoord(50, 50), width=a, height=b,
                                 angle=90. * u.deg)
        assert reg.bounding_box.shape == (a, b)


class TestEllipseSkyRegion(BaseTestSkyRegion):

    reg = EllipseSkyRegion(
        center=SkyCoord(3, 4, unit='deg'),
        width=4 * u.deg,
        height=3 * u.deg,
        angle=5 * u.deg,
    )

    if ASTROPY_LT_13:
        expected_repr = ('<EllipseSkyRegion(<SkyCoord (ICRS): (ra, dec) in '
                    'deg\n    (3.0, 4.0)>, width=4.0 deg, height=3.0 deg,'
                    ' angle=5.0 deg)>')
        expected_str = ('Region: EllipseSkyRegion\ncenter: <SkyCoord (ICRS): '
                   '(ra, dec) in deg\n    (3.0, 4.0)>\nwidth: 4.0 deg\n'
                   'height: 3.0 deg\nangle: 5.0 deg')
    else:
        expected_repr = ('<EllipseSkyRegion(<SkyCoord (ICRS): (ra, dec) in '
                    'deg\n    ( 3.,  4.)>, width=4.0 deg, height=3.0 deg,'
                    ' angle=5.0 deg)>')
        expected_str = ('Region: EllipseSkyRegion\ncenter: <SkyCoord (ICRS): '
                   '(ra, dec) in deg\n    ( 3.,  4.)>\nwidth: 4.0 deg\n'
                   'height: 3.0 deg\nangle: 5.0 deg')

    def test_dimension_center(self):
        center = SkyCoord([1, 2] * u.deg, [3, 4] * u.deg)
        width = 2 * u.arcsec
        height = 3 * u.arcsec
        with pytest.raises(ValueError) as err:
            EllipseSkyRegion(center, width, height)
        assert 'The center must be a 0D SkyCoord object' in str(err)

    def test_contains(self, wcs):
        position = SkyCoord([1, 3] * u.deg, [2, 4] * u.deg)
        # 1,2 is outside, 3,4 is the center and is inside
        assert all(self.reg.contains(position, wcs) == np.array([False, True], dtype='bool'))




import distutils.version as vers
import pytest


import astropy.version as astrov
from astropy.utils.data import get_pkg_data_filename


from ..read import CRTFParser, read_crtf
from ..write import crtf_objects_to_string
from ..core import CRTFRegionParserError

_ASTROPY_MINVERSION = vers.LooseVersion('1.1')
_ASTROPY_VERSION = vers.LooseVersion(astrov.version)

implemented_region_types = ('ellipse', 'circle', 'rectangle', 'poly', 'point', 'text', 'symbol')


def test_global_parser():
    """
    Checks that the global_parser does what's expected.
    """
    global_test_str = str("global coord=B1950_VLA, frame=BARY, corr=[I, Q], color=blue")
    global_parser = CRTFParser(global_test_str)
    assert dict(global_parser.global_meta) == {'coord': 'B1950_VLA', 'frame': 'BARY',
                                               'corr': ['I', 'Q'], 'color': 'blue'}


def test_valid_crtf_line():
    """
    Checks whether a the line is valid CRTF format.
    """
    line_str = 'coord=B1950_VLA, frame=BARY, corr=[I, Q], color=blue'

    with pytest.raises(CRTFRegionParserError) as err:
        CRTFParser(line_str)

    assert 'Not a valid CRTF line:' in str(err)


def test_valid_region_type():
    """
    Checks whether the region type is valid in CRTF format
    """
    reg_str = 'hyperbola[[18h12m24s, -23d11m00s], 2.3arcsec]'

    with pytest.raises(CRTFRegionParserError) as err:
        CRTFParser(reg_str)

    assert "Not a valid CRTF Region type: 'hyperbola'" in str(err)


def test_valid_global_meta_key():
    """
    Checks whether the global key is valid or not.
    """

    global_test_str = str("global label=B1950_VLA, frame=BARY, corr=[I, Q], color=blue")

    with pytest.raises(CRTFRegionParserError) as err:
        CRTFParser(global_test_str)

    assert "'label' is not a valid global meta key" in str(err)


def test_valid_meta_key():
    """
    Checks whether the key is valid or not.
    """

    meta_test_str = str("annulus[[17h51m03.2s, -45d17m50s], [0.10deg, 4.12deg]], hello='My label here'")

    with pytest.raises(CRTFRegionParserError) as err:
        CRTFParser(meta_test_str)

    assert "'hello' is not a valid meta key" in str(err)


def test_valid_region_syntax():
    """
    Checks whether the region has valid parameters
    """

    reg_str1 = "circle[[18h12m24s, -23d11m00s], [2.3arcsec,4.5arcsec]"

    with pytest.raises(CRTFRegionParserError) as err:
        CRTFParser(reg_str1)

    assert "Not in proper format: ('2.3arcsec', '4.5arcsec') should be a single length" in str(err)

    reg_str2 = "symbol[[32.1423deg, 12.1412deg], 12deg], linewidth=2, coord=J2000, symsize=2"

    with pytest.raises(CRTFRegionParserError) as err:
        CRTFParser(reg_str2)

    assert "Not in proper format: '12deg' should be a symbol" in str(err)

    reg_str3 = "circle[[18h12m24s, -23d11m00s]"

    with pytest.raises(CRTFRegionParserError) as err:
        CRTFParser(reg_str3)

    assert "Does not contain expected number of parameters for the region 'circle'" in str(err)

    reg_str4 = "poly[[1, 2], [3,4], [5, 6]]"

    with pytest.raises(CRTFRegionParserError) as err:
        CRTFParser(reg_str4)

    assert "polygon should have > 4 coordinates" in str(err)

    reg_str5 = "poly[[1, 2], [3,4], [5, 6],[1,6]]"

    with pytest.raises(CRTFRegionParserError) as err:
        CRTFParser(reg_str5)

    assert "In polygon, the last and first coordinates should be same" in str(err)

    reg_str6 = "rotbox[[12h01m34.1s, 12d23m33s], [3arcmin,], 12deg]"

    with pytest.raises(CRTFRegionParserError) as err:
        CRTFParser(reg_str6)

    assert "('3arcmin', '') should be a pair of length" in str(err)


@pytest.mark.parametrize('filename', ['data/CRTFgeneral.crtf', 'data/CRTFgeneraloutput.crtf'])
def test_file_crtf(filename):

    filename = get_pkg_data_filename(filename)
    regs = read_crtf(filename, 'warn')
    actual_output = crtf_objects_to_string(regs, 'fk4', '.3f').strip()

    with open(get_pkg_data_filename('data/CRTFgeneraloutput.crtf')) as f:
        ref_output = f.read().strip()

    # since metadata is not required to preserve order, we have to do a more
    # complex comparison
    desired_lines = [set(line.split(",")) for line in ref_output.split("\n")]
    actual_lines = [set(line.split(",")) for line in actual_output.split("\n")]
    for split_line in actual_lines:
        assert split_line in desired_lines

    for split_line in desired_lines:
        assert split_line in actual_lines



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np

from astropy.wcs.utils import skycoord_to_pixel, pixel_to_skycoord

from ..core import PixelRegion, SkyRegion, RegionMask, BoundingBox, PixCoord
from .._geometry import polygonal_overlap_grid
from .._geometry.pnpoly import points_in_polygon
from ..core.attributes import OneDPix, OneDSky, RegionMeta, RegionVisual

__all__ = ['PolygonPixelRegion', 'PolygonSkyRegion']


class PolygonPixelRegion(PixelRegion):
    """
    A polygon in pixel coordinates.

    Parameters
    ----------
    vertices : `~regions.PixCoord`
        The vertices of the polygon
    meta : `~regions.RegionMeta` object, optional
        A dictionary which stores the meta attributes of this region.
    visual : `~regions.RegionVisual` object, optional
        A dictionary which stores the visual meta attributes of this region.

    Examples
    --------

    .. plot::
        :include-source:

        import numpy as np
        from astropy.coordinates import Angle
        from regions import PixCoord, PolygonPixelRegion
        import matplotlib.pyplot as plt

        x, y = [45, 45, 55, 60], [75, 70, 65, 75]
        fig, ax = plt.subplots(1, 1)

        vertices = PixCoord(x=x, y=y)
        reg = PolygonPixelRegion(vertices=vertices)
        patch = reg.as_artist(facecolor='none', edgecolor='red', lw=2)
        ax.add_patch(patch)

        plt.xlim(30, 80)
        plt.ylim(50, 80)
        ax.set_aspect('equal')
        plt.show()
    """

    vertices = OneDPix('vertices')

    def __init__(self, vertices, meta=None, visual=None):
        self.vertices = vertices
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()
        self._repr_params = ('vertices',)

    @property
    def area(self):
        """Region area (float)."""
        raise NotImplementedError

    def contains(self, pixcoord):
        pixcoord = PixCoord._validate(pixcoord, 'pixcoord')
        x = np.atleast_1d(np.asarray(pixcoord.x, dtype=float))
        y = np.atleast_1d(np.asarray(pixcoord.y, dtype=float))
        vx = np.asarray(self.vertices.x, dtype=float)
        vy = np.asarray(self.vertices.y, dtype=float)

        shape = x.shape
        mask = points_in_polygon(x.flatten(), y.flatten(), vx, vy).astype(bool)
        in_poly = mask.reshape(shape)
        if self.meta.get('include', True):
            return in_poly
        else:
            return np.logical_not(in_poly)

    def to_sky(self, wcs):
        vertices_sky = pixel_to_skycoord(self.vertices.x, self.vertices.y, wcs)
        return PolygonSkyRegion(vertices=vertices_sky)

    @property
    def bounding_box(self):
        xmin = self.vertices.x.min()
        xmax = self.vertices.x.max()
        ymin = self.vertices.y.min()
        ymax = self.vertices.y.max()
        return BoundingBox.from_float(xmin, xmax, ymin, ymax)

    def to_mask(self, mode='center', subpixels=5):

        self._validate_mode(mode, subpixels)

        if mode == 'center':
            mode = 'subpixels'
            subpixels = 1

        if mode == 'subpixels':
            use_exact = 0
        else:
            use_exact = 1

        # Find bounding box and mask size
        bbox = self.bounding_box
        ny, nx = bbox.shape

        # Find position of pixel edges and recenter so that circle is at origin
        xmin = float(bbox.ixmin) - 0.5
        xmax = float(bbox.ixmax) - 0.5
        ymin = float(bbox.iymin) - 0.5
        ymax = float(bbox.iymax) - 0.5

        vx = np.asarray(self.vertices.x, dtype=float)
        vy = np.asarray(self.vertices.y, dtype=float)

        fraction = polygonal_overlap_grid(
            xmin, xmax, ymin, ymax,
            nx, ny, vx, vy, use_exact, subpixels,
        )

        return RegionMask(fraction, bbox=bbox)

    def as_artist(self, origin=(0, 0), **kwargs):
        """
        Matplotlib patch object for this region (`matplotlib.patches.Polygon`).

        Parameters:
        -----------
        origin : array_like, optional
            The ``(x, y)`` pixel position of the origin of the displayed image.
            Default is (0, 0).
        kwargs : `dict`
            All keywords that a `~matplotlib.patches.Polygon` object accepts

        Returns
        -------
        patch : `~matplotlib.patches.Polygon`
            Matplotlib polygon patch
        """
        from matplotlib.patches import Polygon
        xy = np.vstack([self.vertices.x - origin[0],
                        self.vertices.y - origin[1]]).transpose()

        mpl_params = self.mpl_properties_default('patch')
        mpl_params.update(kwargs)

        return Polygon(xy=xy, **mpl_params)


class PolygonSkyRegion(SkyRegion):
    """
    A polygon defined using vertices in sky coordinates.

    Parameters
    ----------
    vertices : `~astropy.coordinates.SkyCoord`
        The vertices of the polygon
    meta : `~regions.RegionMeta` object, optional
        A dictionary which stores the meta attributes of this region.
    visual : `~regions.RegionVisual` object, optional
        A dictionary which stores the visual meta attributes of this region.
    """

    vertices = OneDSky('vertices')

    def __init__(self, vertices, meta=None, visual=None):
        self.vertices = vertices
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()
        self._repr_params = ('vertices',)

    def to_pixel(self, wcs):
        x, y = skycoord_to_pixel(self.vertices, wcs)
        vertices_pix = PixCoord(x, y)
        return PolygonPixelRegion(vertices_pix)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import abc
import operator
import inspect

from astropy.extern import six

__all__ = ['Region', 'PixelRegion', 'SkyRegion']


"""
Here we define global variables for the default `origin` and `mode` used
for WCS transformations throughout the `regions` package.

Their purpose is to simplify achieving uniformity across the codebase.
They are mainly used as default arguments for methods that do WCS
transformations.

They are private (with an underscore), not part of the public API,
users should not touch them.
"""
_DEFAULT_WCS_ORIGIN = 0
_DEFAULT_WCS_MODE = 'all'

VALID_MASK_MODES = {'center', 'exact', 'subpixels'}


class MetaRegion(abc.ABCMeta):
    """
    This is a sub metaclass of `abc.ABCMeta` that makes methods of a class
    automatically have their docstrings filled in from the methods they override
    in the base class.

    If the class uses multiple inheritance, the docstring will be
    chosen from the first class in the bases list, in the same way as
    methods are normally resolved in Python.  If this results in
    selecting the wrong docstring, the docstring will need to be
    explicitly included on the method.

    """

    def __init__(cls, name, bases, dct):
        def is_public_member(key):
            return (
                (key.startswith('__') and key.endswith('__')
                 and len(key) > 4) or
                not key.startswith('_'))

        for key, val in six.iteritems(dct):
            if (inspect.isfunction(val) and is_public_member(key)
                    and val.__doc__ is None):
                for base in cls.__mro__[1:]:
                    super_method = getattr(base, key, None)
                    if super_method is not None:
                        val.__doc__ = super_method.__doc__
                        break

        super(MetaRegion, cls).__init__(name, bases, dct)


@six.add_metaclass(MetaRegion)
class Region(object):
    """
    Base class for all regions.
    """

    def __repr__(self):
        if hasattr(self, 'center'):
            params = [repr(self.center)]
        else:
            params = []
        if self._repr_params is not None:
            for key in self._repr_params:
                params.append('{0}={1}'.format(key.replace("_", " "),
                                               getattr(self, key)))
        params = ', '.join(params)

        return '<{0}({1})>'.format(self.__class__.__name__, params)

    def __str__(self):
        cls_info = [('Region', self.__class__.__name__)]
        if hasattr(self, 'center'):
            cls_info.append(('center', self.center))
        if self._repr_params is not None:
            params_value = [(x.replace("_", " "), getattr(self, x))
                            for x in self._repr_params]
            cls_info += params_value
        fmt = ['{0}: {1}'.format(key, val) for key, val in cls_info]

        return '\n'.join(fmt)

    @abc.abstractmethod
    def intersection(self, other):
        """
        Returns a region representing the intersection of this region with
        ``other``.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def symmetric_difference(self, other):
        """
        Returns the union of the two regions minus any areas contained in the
        intersection of the two regions.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def union(self, other):
        """
        Returns a region representing the union of this region with ``other``.
        """
        raise NotImplementedError

    def __and__(self, other):
        return self.intersection(other)

    def __or__(self, other):
        return self.union(other)

    def __xor__(self, other):
        return self.symmetric_difference(other)


class PixelRegion(Region):
    """
    Base class for all regions defined in pixel coordinates
    """

    def intersection(self, other):
        """
        Returns a region representing the intersection of this region with
        ``other``.
        """
        from .compound import CompoundPixelRegion
        return CompoundPixelRegion(region1=self, region2=other, operator=operator.and_)

    def symmetric_difference(self, other):
        """
        Returns the union of the two regions minus any areas contained in the
        intersection of the two regions.
        """
        from .compound import CompoundPixelRegion
        return CompoundPixelRegion(region1=self, region2=other, operator=operator.xor)

    def union(self, other):
        """
        Returns a region representing the union of this region with ``other``.
        """
        from .compound import CompoundPixelRegion
        return CompoundPixelRegion(region1=self, region2=other, operator=operator.or_)

    @abc.abstractmethod
    def contains(self, pixcoord):
        """
        Checks whether a position or positions fall inside the region.

        Parameters
        ----------
        pixcoord : `~regions.PixCoord`
            The position or positions to check.
        """
        raise NotImplementedError

    def __contains__(self, coord):
        if not coord.isscalar:
            raise ValueError('coord must be scalar. coord={}'.format(coord))
        return self.contains(coord)

    @abc.abstractmethod
    def to_sky(self, wcs):
        """
        Returns a region defined in sky coordinates.

        Parameters
        ----------
        wcs : `~astropy.wcs.WCS` instance
            The world coordinate system transformation to assume

        Returns
        -------
        sky_region : `~regions.SkyRegion` object.
        """
        raise NotImplementedError

    @abc.abstractproperty
    def bounding_box(self):
        """
        The minimal bounding box (in integer pixel coordinates) that contains
        the region.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def to_mask(self, mode='center', subpixels=5):
        """
        Returns a mask for the aperture.

        Parameters
        ----------
        mode : { 'center' | 'exact' | 'subpixels'}, optional
            The following modes are available:
                * ``'center'``: returns 1 for pixels where the center is in
                  the region, and 0 otherwise.
                * ``'exact'``: returns a value between 0 and 1 giving the
                  fractional level of overlap of the pixel with the region.
                * ``'subpixels'``: A pixel is divided into subpixels and
                  the center of each subpixel is tested (a subpixel is
                  either completely in or out of the region).  Returns a
                  value between 0 and 1 giving the fractional level of
                  overlap of the subpixels with the region.  With
                  ``subpixels`` set to 1, this method is equivalent to
                  ``'center'``.
        subpixels : int, optional
            For the ``'subpixel'`` mode, resample pixels by this factor
            in each dimension. That is, each pixel is divided into
            ``subpixels ** 2`` subpixels.

        Returns
        -------
        mask : `~regions.Mask`
            A region mask object.
        """
        raise NotImplementedError

    @staticmethod
    def _validate_mode(mode, subpixels):
        if mode not in VALID_MASK_MODES:
            raise ValueError("Invalid mask mode: {0} (should be one "
                             "of {1})".format(mode, '/'.join(VALID_MASK_MODES)))
        if mode == 'subpixels':
            if not isinstance(subpixels, int) or subpixels <= 0:
                raise ValueError("Invalid subpixels value: {0} (should be"
                                 " a strictly positive integer)".format(subpixels))

    @abc.abstractmethod
    def as_artist(self, origin=(0, 0), **kwargs):
        """
        Convert to mpl patch

        Parameters
        ----------
        origin : array_like, optional
            The ``(x, y)`` pixel position of the origin of the displayed image.
            Default is (0, 0).

        kwargs : `dict`
            keywords that a `~matplotlib.patches.Patch` accepts

        Returns
        -------
        patch : `~matplotlib.patches.Patch`
            Matplotlib patch
        """
        raise NotImplementedError

    def mpl_properties_default(self, shape='patch'):
        """
        This sets the default values of the visual attributes as specified
        under DS9 convention.
        """

        kwargs = dict()
        kwargs['color'] = self.visual.get('color', 'green')
        kwargs['label'] = self.meta.get('label', "")

        if shape == 'text':
            kwargs['family'] = self.visual.get('font', 'helvetica')
            kwargs['rotation'] = self.visual.get('textangle', '0')
            kwargs['size'] = self.visual.get('fontsize', '12')
            kwargs['style'] = self.visual.get('fontsyle', 'normal')
            kwargs['weight'] = self.visual.get('fontweight', 'roman')

        else:
            if shape == 'Line2D':
                kwargs['marker'] = self.visual.get('symbol', 'o')
                kwargs['markersize'] = self.visual.get('symsize', 11)
                kwargs['markeredgecolor'] = kwargs['color']
                kwargs['markeredgewidth'] = self.visual.get('width', 1)
            if shape == 'patch':
                kwargs['edgecolor'] = kwargs.pop('color')
                kwargs['fill'] = self.visual.get('fill', False)

            kwargs['linewidth'] = self.visual.get('linewidth', 1)
            kwargs['linestyle'] = self.visual.get('linstyle', 'solid')

        return kwargs

    def plot(self, origin=(0, 0), ax=None, **kwargs):
        """
        Calls ``as_artist`` method forwarding all kwargs and adds patch
        to given axis.

        Parameters
        ----------
        origin : array_like, optional
            The ``(x, y)`` pixel position of the origin of the displayed image.
            Default is (0, 0).
        ax : `~matplotlib.axes.Axes`, optional
            Axis
        kwargs : `dict`
            keywords that a `~matplotlib.patches.Patch` accepts

        Returns
        -------
        ax : `~matplotlib.axes.Axes`
            Axes on which the patch is added.
        """
        import matplotlib.pyplot as plt

        if ax is None:
            ax = plt.gca()

        artist = self.as_artist(origin=origin, **kwargs)
        ax.add_artist(artist)

        return ax


class SkyRegion(Region):
    """
    Base class for all regions defined in celestial coordinates
    """

    def intersection(self, other):
        """
        Returns a region representing the intersection of this region with
        ``other``.
        """
        from .compound import CompoundSkyRegion
        return CompoundSkyRegion(region1=self, region2=other, operator=operator.and_)

    def symmetric_difference(self, other):
        """
        Returns the union of the two regions minus any areas contained in the
        intersection of the two regions.
        """
        from .compound import CompoundSkyRegion
        return CompoundSkyRegion(region1=self, region2=other, operator=operator.xor)

    def union(self, other):
        """
        Returns a region representing the union of this region with ``other``.
        """
        from .compound import CompoundSkyRegion
        return CompoundSkyRegion(region1=self, region2=other, operator=operator.or_)

    def contains(self, skycoord, wcs):
        """
        Check whether a sky coordinate falls inside the region

        Parameters
        ----------
        skycoord : `~astropy.coordinates.SkyCoord`
            The position or positions to check
        wcs : `~astropy.wcs.WCS` instance
            The world coordinate system transformation to assume
        """
        from .pixcoord import PixCoord
        pixel_region = self.to_pixel(wcs)
        pixcoord = PixCoord.from_sky(skycoord, wcs)
        return pixel_region.contains(pixcoord)

    @abc.abstractmethod
    def to_pixel(self, wcs):
        """
        Returns the equivalent region defined in pixel coordinates.

        Parameters
        ----------
        wcs : `~astropy.wcs.WCS` instance
            The world coordinate system transformation to assume

        Returns
        -------
        pixel_region : `~regions.PixelRegion` object.
        """
        raise NotImplementedError



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals

from ..core import to_shape_list

__all__ = [
    'write_crtf',
    'crtf_objects_to_string',
]


def crtf_objects_to_string(regions, coordsys='fk5', fmt='.6f', radunit='deg'):
    """
    Converts a `list` of `~regions.Region` to CRTF region string.

    Parameters
    ----------
    regions : `list`
        List of `~regions.Region` objects
    coordsys : `str`, optional
        Astropy Coordinate system that overrides the coordinate system frame for
        all regions. Default is 'fk5'.
    fmt : `str`, optional
        A python string format defining the output precision. Default is .6f,
        which is accurate to 0.0036 arcseconds.
    radunit : `str`, optional
        This denotes the unit of the radius. Default is deg (degrees)

    Returns
    -------
    region_string : `str`
        CRTF region string

    Examples
    --------
    >>> from astropy import units as u
    >>> from astropy.coordinates import SkyCoord
    >>> from regions import CircleSkyRegion, crtf_objects_to_string
    >>> reg_sky = CircleSkyRegion(SkyCoord(1 * u.deg, 2 * u.deg), 5 * u.deg)
    >>> print(crtf_objects_to_string([reg_sky]))
    #CRTF
    global coord=fk5
    +circle[[1.000007deg, 2.000002deg], 5.000000deg]

    """

    shapelist = to_shape_list(regions, coordsys)
    return shapelist.to_crtf(coordsys, fmt, radunit)


def write_crtf(regions, filename, coordsys='fk5', fmt='.6f', radunit='deg'):
    """
    Converts a `list` of `~regions.Region` to CRTF string and write to file.

    Parameters
    ----------
    regions : `list`
        List of `~regions.Region` objects
    filename : `str`
        Filename in which the string is to be written. Default is 'new.crtf'
    coordsys : `str`, optional
        Astropy Coordinate system that overrides the coordinate frames of all
        regions. Default is 'fk5'.
    fmt : `str`, optional
        A python string format defining the output precision. Default is .6f,
        which is accurate to 0.0036 arcseconds.
    radunit : `str`, optional
        This denotes the unit of the radius. Default is deg (degrees)

    Examples
    --------
    >>> from astropy import units as u
    >>> from astropy.coordinates import SkyCoord
    >>> from regions import CircleSkyRegion, write_crtf
    >>> reg_sky = CircleSkyRegion(SkyCoord(1 * u.deg, 2 * u.deg), 5 * u.deg)
    >>> write_crtf([reg_sky], 'test_write.crtf')
    >>> with open('test_write.crtf') as f:
    ...      print(f.read())
    #CRTF
    global coord=fk5
    +circle[[1.000007deg, 2.000002deg], 5.000000deg]
    """

    output = crtf_objects_to_string(regions, coordsys, fmt, radunit)
    with open(filename, 'w') as fh:
        fh.write(output)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Region core functionality.
"""
from .core import *
from .pixcoord import *
from .mask import *
from .bounding_box import *
from .compound import *
from .attributes import *



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function
import string
import itertools
import re
import copy
from collections import OrderedDict
from warnings import warn

from astropy import units as u
from astropy import coordinates
from astropy import log

from ..core import reg_mapping
from ..core import Shape, ShapeList
from .core import DS9RegionParserError, DS9RegionParserWarning, valid_symbols_ds9

__all__ = [
    'read_ds9',
    'DS9Parser',
    'DS9RegionParser',
    'CoordinateParser'
]

# Regular expression to extract region type or coodinate system
regex_global = re.compile("^#? *(-?)([a-zA-Z0-9]+)")

# Regular expression to extract meta attributes
regex_meta = re.compile("([a-zA-Z]+)(=)({.*?}|\'.*?\'|\".*?\"|[0-9\s]+\s?|[^=\s]+\s?[0-9]*)\s?")

# Regular expression to strip parenthesis
regex_paren = re.compile("[()]")

# Regular expression to split coordinate strings
regex_splitter = re.compile("[, ]")


def read_ds9(filename, errors='strict'):
    """
    Read a DS9 region file in as a `list` of `~regions.Region` objects.

    Parameters
    ----------
    filename : `str`
        The file path
    errors : ``warn``, ``ignore``, ``strict``, optional
      The error handling scheme to use for handling parsing errors.
      The default is 'strict', which will raise a `~regions.DS9RegionParserError`.
      ``warn`` will raise a `~regions.DS9RegionParserWarning`, and
      ``ignore`` will do nothing (i.e., be silent).

    Returns
    -------
    regions : `list`
        Python list of `~regions.Region` objects.

    Examples
    --------
    >>> from regions import read_ds9
    >>> from astropy.utils.data import get_pkg_data_filename
    >>> file = get_pkg_data_filename('data/physical_reference.reg', package='regions.io.ds9.tests')
    >>> regs = read_ds9(file, errors='warn')
    >>> print(regs[0])
    Region: CirclePixelRegion
    center: PixCoord(x=330.0, y=1090.0)
    radius: 40.0
    >>> print(regs[0].meta)
    {'label': 'Circle', 'select': '1', 'highlite': '1', 'fixed': '0', 'edit': '1', 'move': '1', 'delete': '1', 'source': '1', 'tag': ['{foo}', '{foo bar}'], 'include': True}
    >>> print(regs[0].visual)
    {'dashlist': '8 3', 'dash': '0', 'color': 'pink', 'linewidth': '3', 'font': 'times', 'fontsize': '10', 'fontstyle': 'normal', 'fontweight': 'roman'}

    """
    with open(filename) as fh:
        region_string = fh.read()

    parser = DS9Parser(region_string, errors=errors)
    return parser.shapes.to_regions()


class CoordinateParser(object):
    """
    Helper class to structure coordinate parser
    """
    @staticmethod
    def parse_coordinate(string_rep, unit):
        """
        Parse a single coordinate
        """
        # explicit radian ('r') value
        if string_rep[-1] == 'r':
            return coordinates.Angle(string_rep[:-1], unit=u.rad)
        # explicit image ('i') and physical ('p') pixels
        elif string_rep[-1] in ['i', 'p']:
            return u.Quantity(string_rep[:-1]) - 1
        # Any ds9 coordinate representation (sexagesimal or degrees)
        elif 'd' in string_rep or 'h' in string_rep:
            return coordinates.Angle(string_rep)
        elif unit is 'hour_or_deg':
            if ':' in string_rep:
                spl = tuple([float(x) for x in string_rep.split(":")])
                return coordinates.Angle(spl, u.hourangle)
            else:
                ang = float(string_rep)
                return coordinates.Angle(ang, u.deg)
        elif unit.is_equivalent(u.deg):
            # return coordinates.Angle(string_rep, unit=unit)
            if ':' in string_rep:
                ang = tuple([float(x) for x in string_rep.split(":")])
            else:
                ang = float(string_rep)
            return coordinates.Angle(ang, u.deg)
        elif unit.is_equivalent(u.dimensionless_unscaled):
            return u.Quantity(float(string_rep), unit) - 1
        else:
            return u.Quantity(float(string_rep), unit)

    @staticmethod
    def parse_angular_length_quantity(string_rep, unit=u.deg):
        """
        Given a string that is either a number or a number and a unit, return a
        Quantity of that string.  e.g.:

            23.9 -> 23.9*u.deg
            50" -> 50*u.arcsec
        """
        unit_mapping = {
            '"': u.arcsec,
            "'": u.arcmin,
            'd': u.deg,
            'r': u.rad,
            'i': u.dimensionless_unscaled,
            'p': u.dimensionless_unscaled
        }
        has_unit = string_rep[-1] not in string.digits
        if has_unit:
            unit = unit_mapping[string_rep[-1]]
            return u.Quantity(float(string_rep[:-1]), unit=unit)
        else:
            return u.Quantity(float(string_rep), unit=unit)


class DS9Parser(object):
    """
    Parse a DS9 string

    This class transforms a DS9 string to a `~regions.io.core.ShapeList`. The
    result is stored as ``shapes`` attribute.

    Each line is tested for either containing a region type or a coordinate
    system. If a coordinate system is found the global coordsys state of the
    parser is modified. If a region type is found the
    `~regions.DS9RegionParser` is invokes to transform the line into a
    `~regions.Shape` object.

    Parameters
    ----------
    region_string : `str`
        DS9 region string
    errors : ``warn``, ``ignore``, ``strict``, optional
      The error handling scheme to use for handling parsing errors.
      The default is 'strict', which will raise a `~regions.DS9RegionParserError`.
      ``warn`` will raise a `~regions.DS9RegionParserWarning`, and
      ``ignore`` will do nothing (i.e., be silent).

    Examples
    --------
    >>> from regions import DS9Parser
    >>> reg_str = 'image circle(331.00,1091.00,40.00) # dashlist=8 3 select=1 highlite=1 dash=0 fixed=0 edit=1 move=1 delete=1 source=1 text={Circle} tag={foo} tag={foo bar} This is a Comment color=pink width=3 font="times 10 normal roman"'
    >>> regs = DS9Parser(reg_str, errors='warn').shapes.to_regions()
    >>> print(regs[0])
    Region: CirclePixelRegion
    center: PixCoord(x=330.0, y=1090.0)
    radius: 40.0
    >>> print(regs[0].meta)
    {'label': 'Circle', 'select': '1', 'highlite': '1', 'fixed': '0', 'edit': '1', 'move': '1', 'delete': '1', 'source': '1', 'tag': ['{foo}', '{foo bar}'], 'include': True}
    >>> print(regs[0].visual)
    {'dashlist': '8 3', 'dash': '0', 'color': 'pink', 'linewidth': '3', 'font': 'times', 'fontsize': '10', 'fontstyle': 'normal', 'fontweight': 'roman'}
    """

    # List of valid coordinate system
    coordinate_systems = ['fk5', 'fk4', 'icrs', 'galactic', 'wcs', 'physical', 'image', 'ecliptic', 'J2000']
    coordinate_systems += ['wcs{0}'.format(letter) for letter in string.ascii_lowercase]

    # Map to convert coordinate system names
    coordsys_mapping = dict(zip(coordinates.frame_transform_graph.get_names(),
                                coordinates.frame_transform_graph.get_names()))
    coordsys_mapping['ecliptic'] = 'geocentrictrueecliptic'
    coordsys_mapping['J2000'] = 'fk5'

    def __init__(self, region_string, errors='strict'):
        if errors not in ('strict', 'ignore', 'warn'):
            msg = "``errors`` must be one of strict, ignore, or warn; is {}"
            raise ValueError(msg.format(errors))
        self.region_string = region_string
        self.errors = errors

        # Global states
        self.coordsys = None
        self.global_meta = {}

        # Results
        self.shapes = ShapeList()

        self.run()

    def __str__(self):
        ss = self.__class__.__name__
        ss += '\nErrors: {}'.format(self.errors)
        ss += '\nCoordsys: {}'.format(self.coordsys)
        ss += '\nGlobal meta: {}'.format(self.global_meta)
        ss += '\nShapes: {}'.format(self.shapes)
        ss += '\n'
        return ss

    def set_coordsys(self, coordsys):
        """
        Transform coordinate system

        # TODO: needs expert attention
        """
        if coordsys in self.coordsys_mapping:
            self.coordsys = self.coordsys_mapping[coordsys]
        else:
            self.coordsys = coordsys

    def run(self):
        """
        Run all steps
        """
        for line_ in self.region_string.split('\n'):
            for line in line_.split(";"):
                self.parse_line(line)
                log.debug('Global state: {}'.format(self))

    def parse_line(self, line):
        """
        Parse one line
        """
        log.debug('Parsing {}'.format(line))

        # Skip blanks
        if line == '':
            return

        # Skip comments
        if line[0] == '#':
            return

        # Special case / header: parse global parameters into metadata
        if line.lstrip()[:6] == 'global':
            self.global_meta = self.parse_meta(line)
            # global_meta can specify "include=1"; never seen other options
            # used but presumably =0 means false
            self.global_meta['include'] = (False if
                                           self.global_meta.get('include') in
                                           ('0', 'False', False) else True)
            return

        # Try to parse the line
        region_type_search = regex_global.search(line)
        if region_type_search:
            include = region_type_search.groups()[0]
            region_type = region_type_search.groups()[1]
        else:
            self._raise_error("No region type found for line '{0}'.".format(line))
            return

        if region_type in self.coordinate_systems:
            # Found coord system definition
            self.set_coordsys(region_type)
            return
        if region_type not in DS9RegionParser.language_spec:
            self._raise_error("Region type '{0}' was identified, but it is not one of "
                              "the known region types.".format(region_type))
            return
        else:
            # Found region specification,
            region_end = region_type_search.span()[1]
            self.parse_region(include, region_type, region_end, line)

    def _raise_error(self, msg):
        if self.errors == 'warn':
            warn(msg, DS9RegionParserWarning)
        elif self.errors == 'strict':
            raise DS9RegionParserError(msg)

    @staticmethod
    def parse_meta(meta_str):
        """
        Parse the metadata for a single ds9 region string.

        Parameters
        ----------
        meta_str : `str`
            Meta string, the metadata is everything after the close-paren of the
            region coordinate specification. All metadata is specified as
            key=value pairs separated by whitespace, but sometimes the values
            can also be whitespace separated.

        Returns
        -------
        meta : `~collections.OrderedDict`
            Dictionary containing the meta data
        """
        keys_vals = [(x, y) for x, _, y in regex_meta.findall(meta_str.strip())]
        extra_text = regex_meta.split(meta_str.strip())[-1]
        result = OrderedDict()
        for key, val in keys_vals:
            # regex can include trailing whitespace or inverted commas
            # remove it
            val = val.strip().strip("'").strip('"')
            if key == 'text':
                val = val.lstrip("{").rstrip("}")
            if key in result:
                if key == 'tag':
                    result[key].append(val)
                else:
                    raise ValueError("Duplicate key {0} found".format(key))
            else:
                if key == 'tag':
                    result[key] = [val]
                else:
                    result[key] = val
        if extra_text:
            result['comment'] = extra_text

        return result

    def parse_region(self, include, region_type, region_end, line):
        """
        Extract a Shape from a region string
        """
        if self.coordsys is None:
            raise DS9RegionParserError("No coordinate system specified and a"
                                       " region has been found.")
        else:
            helper = DS9RegionParser(coordsys=self.coordsys,
                                     include=include,
                                     region_type=region_type,
                                     region_end=region_end,
                                     global_meta=self.global_meta,
                                     line=line)
            helper.parse()
            self.shapes.append(helper.shape)


# Global definitions to improve readability
radius = CoordinateParser.parse_angular_length_quantity
width = CoordinateParser.parse_angular_length_quantity
height = CoordinateParser.parse_angular_length_quantity
angle = CoordinateParser.parse_angular_length_quantity
coordinate = CoordinateParser.parse_coordinate


class DS9RegionParser(object):
    """
    Parse a DS9 region string

    This will turn a line containing a DS9 region into a Shape

    Parameters
    ----------
    coordsys : `str`
        Coordinate system
    include : `str` {'', '-'}
        Flag at the beginning of the line
    region_type : `str`
        Region type
    region_end : `int`
        Coordinate of the end of the regions name, this is passed in order to
        handle whitespaces correctly
    global_meta : `dict`
        Global meta data
    line : `str`
        Line to parse
    """

    # Coordinate unit transformations
    coordinate_units = {'fk5': ('hour_or_deg', u.deg),
                        'fk4': ('hour_or_deg', u.deg),
                        'icrs': ('hour_or_deg', u.deg),
                        'geocentrictrueecliptic': (u.deg, u.deg),
                        'galactic': (u.deg, u.deg),
                        'physical': (u.dimensionless_unscaled, u.dimensionless_unscaled),
                        'image': (u.dimensionless_unscaled, u.dimensionless_unscaled),
                        'wcs': (u.dimensionless_unscaled, u.dimensionless_unscaled),
                        }
    for letter in string.ascii_lowercase:
        coordinate_units['wcs{0}'.format(letter)] = (u.dimensionless_unscaled, u.dimensionless_unscaled)

    # DS9 language specification. This defines how a certain region is read.
    language_spec = {'point': (coordinate, coordinate),
                     'text': (coordinate, coordinate),
                     'circle': (coordinate, coordinate, radius),
                     # This is a special case to deal with n elliptical annuli
                     'ellipse': itertools.chain((coordinate, coordinate),
                                                itertools.cycle((radius,))),
                     'box': (coordinate, coordinate, width, height, angle),
                     'polygon': itertools.cycle((coordinate,)),
                     'line': (coordinate, coordinate, coordinate, coordinate),
                     'annulus': itertools.chain((coordinate, coordinate),
                                                itertools.cycle((radius,))),
                     }

    def __init__(self, coordsys, include, region_type, region_end, global_meta, line):

        self.coordsys = coordsys
        self.include = include
        self.region_type = region_type
        self.region_end = region_end
        self.global_meta = global_meta
        self.line = line

        self.meta_str = None
        self.coord_str = None
        self.composite = None
        self.coord = None
        self.meta = None
        self.shape = None

    def __str__(self):
        ss = self.__class__.__name__
        ss += '\nLine : {}'.format(self.line)
        ss += '\nRegion end : {}'.format(self.region_end)
        ss += '\nMeta string : {}'.format(self.meta_str)
        ss += '\nCoord string: {}'.format(self.coord_str)
        ss += '\nShape: {}'.format(self.shape)
        ss += '\n'
        return ss

    def parse(self):
        """
        Convert line to shape object
        """
        log.debug(self)

        self.parse_composite()
        self.split_line()
        self.convert_coordinates()
        self.convert_meta()
        self.make_shape()
        log.debug(self)

    def parse_composite(self):
        """
        Determine whether the region is composite
        """
        self.composite = "||" in self.line

    def split_line(self):
        """
        Split line into coordinates and meta string
        """
        # coordinate of the # symbol or end of the line (-1) if not found
        hash_or_end = self.line.find("#")
        temp = self.line[self.region_end:hash_or_end].strip(" |")
        self.coord_str = regex_paren.sub("", temp)

        # don't want any meta_str if there is no metadata found
        if hash_or_end >= 0:
            self.meta_str = self.line[hash_or_end:]
        else:
            self.meta_str = ""

    def convert_coordinates(self):
        """
        Convert coordinate string to objects
        """
        coord_list = []
        # strip out "null" elements, i.e. ''.  It might be possible to eliminate
        # these some other way, i.e. with regex directly, but I don't know how.
        # We need to copy in order not to burn up the iterators
        elements = [x for x in regex_splitter.split(self.coord_str) if x]
        element_parsers = self.language_spec[self.region_type]
        for ii, (element, element_parser) in enumerate(zip(elements,
                                                           element_parsers)):
            if element_parser is coordinate:
                unit = self.coordinate_units[self.coordsys][ii % 2]
                coord_list.append(element_parser(element, unit))
            elif self.coordinate_units[self.coordsys][0] is u.dimensionless_unscaled:
                    coord_list.append(element_parser(element, unit=u.dimensionless_unscaled))
            else:
                coord_list.append(element_parser(element))

        if self.region_type in ['ellipse', 'box'] and len(coord_list) % 2 == 1:
                coord_list[-1] = CoordinateParser.parse_angular_length_quantity(elements[len(coord_list)-1])

        # Reset iterator for ellipse and annulus
        # Note that this cannot be done with copy.deepcopy on python2
        if self.region_type in ['ellipse', 'annulus']:
            self.language_spec[self.region_type] = itertools.chain(
                (coordinate, coordinate), itertools.cycle((radius,)))

        self.coord = coord_list

    def convert_meta(self):
        """
        Convert meta string to dict
        """
        meta_ = DS9Parser.parse_meta(self.meta_str)
        self.meta = copy.deepcopy(self.global_meta)
        self.meta.update(meta_)
        # the 'include' is not part of the metadata string;
        # it is pre-parsed as part of the shape type and should always
        # override the global one
        self.include = self.meta.get('include', True) if self.include == '' else self.include != '-'
        self.meta['include'] = self.include

    def make_shape(self):
        """
        Make shape object
        """
        # In DS9, ellipse can also represents an elliptical annulus
        # For elliptical annulus angle is optional.
        if self.region_type == 'ellipse':
            self.coord[2:] = [x * 2 for x in self.coord[2:]]
            if len(self.coord) % 2 == 1:  # This checks if angle is present
                self.coord[-1] /= 2

        if 'point' in self.meta:
            point = self.meta['point'].split(" ")
            if len(point) > 1:
                self.meta['symsize'] = point[1]
            self.meta['point'] = valid_symbols_ds9[point[0]]

        if 'font' in self.meta:
            fonts = self.meta['font'].split(" ")
            keys = ['font', 'fontsize', 'fontstyle', 'fontweight']
            for i, val in enumerate(fonts):
                self.meta[keys[i]] = val

        self.meta.pop('coord', None)

        self.shape = Shape(coordsys=self.coordsys,
                           region_type=reg_mapping['DS9'][self.region_type],
                           coord=self.coord,
                           meta=self.meta,
                           composite=self.composite,
                           include=self.include,
                          )



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
from astropy import units as u

from astropy.coordinates import Angle
from astropy.wcs.utils import pixel_to_skycoord

from ..core import PixCoord, PixelRegion, SkyRegion, RegionMask, BoundingBox
from .._geometry import rectangular_overlap_grid
from .._utils.wcs_helpers import skycoord_to_pixel_scale_angle
from ..core.attributes import ScalarPix, ScalarLength, QuantityLength, ScalarSky
from .polygon import PolygonPixelRegion


__all__ = ['RectanglePixelRegion', 'RectangleSkyRegion']


class RectanglePixelRegion(PixelRegion):
    """
    A rectangle in pixel coordinates.

    Parameters
    ----------
    center : `~regions.PixCoord`
        The position of the center of the rectangle.
    width : `float`
        The width of the rectangle (before rotation) in pixels
    height : `float`
        The height of the rectangle (before rotation) in pixels
    angle : `~astropy.units.Quantity`, optional
        The rotation angle of the rectangle, measured anti-clockwise. If set to
        zero (the default), the width axis is lined up with the x axis.
    meta : `~regions.RegionMeta` object, optional
        A dictionary which stores the meta attributes of this region.
    visual : `~regions.RegionVisual` object, optional
        A dictionary which stores the visual meta attributes of this region.

    Examples
    --------

    .. plot::
        :include-source:

        import numpy as np
        from astropy.coordinates import Angle
        from regions import PixCoord, RectanglePixelRegion
        import matplotlib.pyplot as plt

        x, y = 15, 10
        width, height = 8, 5
        angle = Angle(30, 'deg')

        fig, ax = plt.subplots(1, 1)

        center = PixCoord(x=x, y=y)
        reg = RectanglePixelRegion(center=center, width=width,
                                   height=height, angle=angle)
        patch = reg.as_artist(facecolor='none', edgecolor='red', lw=2)
        ax.add_patch(patch)

        plt.xlim(0, 30)
        plt.ylim(0, 20)
        ax.set_aspect('equal')
        plt.show()
    """

    center = ScalarPix('center')
    width = ScalarLength('width')
    height = ScalarLength('height')
    angle = QuantityLength('angle')

    def __init__(self, center, width, height, angle=0 * u.deg, meta=None, visual=None):
        self.center = center
        self.width = width
        self.height = height
        self.angle = angle
        self.meta = meta or {}
        self.visual = visual or {}
        self._repr_params = ('width', 'height', 'angle')

    @property
    def area(self):
        """Region area (float)"""
        return self.width * self.height

    def contains(self, pixcoord):
        cos_angle = np.cos(self.angle)
        sin_angle = np.sin(self.angle)
        dx = pixcoord.x - self.center.x
        dy = pixcoord.y - self.center.y
        dx_rot = cos_angle * dx + sin_angle * dy
        dy_rot = sin_angle * dx - cos_angle * dy
        in_rect = (np.abs(dx_rot) < self.width * 0.5) & (np.abs(dy_rot) < self.height * 0.5)
        if self.meta.get('include', True):
            return in_rect
        else:
            return np.logical_not(in_rect)

    def to_sky(self, wcs):
        # TODO: write a pixel_to_skycoord_scale_angle
        center = pixel_to_skycoord(self.center.x, self.center.y, wcs)
        _, scale, north_angle = skycoord_to_pixel_scale_angle(center, wcs)
        width = Angle(self.width / scale, 'deg')
        height = Angle(self.height / scale, 'deg')
        return RectangleSkyRegion(center, width, height,
                                  angle=self.angle - (north_angle - 90 * u.deg),
                                  meta=self.meta, visual=self.visual)

    @property
    def bounding_box(self):
        """
        The minimal bounding box (`~regions.BoundingBox`) enclosing the
        exact rectangular region.
        """

        w2 = self.width / 2.
        h2 = self.height / 2.
        cos_angle = np.cos(self.angle)    # self.angle is a Quantity
        sin_angle = np.sin(self.angle)    # self.angle is a Quantity
        dx1 = abs(w2 * cos_angle - h2 * sin_angle)
        dy1 = abs(w2 * sin_angle + h2 * cos_angle)
        dx2 = abs(w2 * cos_angle + h2 * sin_angle)
        dy2 = abs(w2 * sin_angle - h2 * cos_angle)
        dx = max(dx1, dx2)
        dy = max(dy1, dy2)

        xmin = self.center.x - dx
        xmax = self.center.x + dx
        ymin = self.center.y - dy
        ymax = self.center.y + dy

        return BoundingBox.from_float(xmin, xmax, ymin, ymax)

    def to_mask(self, mode='center', subpixels=5):

        # NOTE: assumes this class represents a single circle

        self._validate_mode(mode, subpixels)

        if mode == 'center':
            mode = 'subpixels'
            subpixels = 1

        # Find bounding box and mask size
        bbox = self.bounding_box
        ny, nx = bbox.shape

        # Find position of pixel edges and recenter so that circle is at origin
        xmin = float(bbox.ixmin) - 0.5 - self.center.x
        xmax = float(bbox.ixmax) - 0.5 - self.center.x
        ymin = float(bbox.iymin) - 0.5 - self.center.y
        ymax = float(bbox.iymax) - 0.5 - self.center.y

        if mode == 'subpixels':
            use_exact = 0
        else:
            use_exact = 1

        fraction = rectangular_overlap_grid(
            xmin, xmax, ymin, ymax, nx, ny,
            self.width, self.height,
            self.angle.to(u.rad).value,
            use_exact, subpixels,
        )

        return RegionMask(fraction, bbox=bbox)

    def as_artist(self, origin=(0, 0), **kwargs):
        """
        Matplotlib patch object for this region (`matplotlib.patches.Rectangle`).

        Parameters:
        -----------
        origin : array_like, optional
            The ``(x, y)`` pixel position of the origin of the displayed image.
            Default is (0, 0).
        kwargs : `dict`
            All keywords that a `~matplotlib.patches.Rectangle` object accepts

        Returns
        -------
        patch : `~matplotlib.patches.Rectangle`
            Matplotlib circle patch
        """
        from matplotlib.patches import Rectangle
        xy = self._lower_left_xy()
        xy = xy[0] - origin[0], xy[1] - origin[1]
        width = self.width
        height = self.height
        # From the docstring: MPL expects "rotation in degrees (anti-clockwise)"
        angle = self.angle.to('deg').value

        mpl_params = self.mpl_properties_default('patch')
        mpl_params.update(kwargs)

        return Rectangle(xy=xy, width=width, height=height,
                         angle=angle, **mpl_params)

    @property
    def corners(self):
        """
        Return the x, y coordinate pairs that define the corners
        """

        corners = [(-self.width/2, -self.height/2),
                   ( self.width/2, -self.height/2),
                   ( self.width/2,  self.height/2),
                   (-self.width/2,  self.height/2),
                  ]
        rotmat = [[np.cos(self.angle), np.sin(self.angle)],
                  [-np.sin(self.angle), np.cos(self.angle)]]

        return np.dot(corners, rotmat) + np.array([self.center.x,
                                                   self.center.y])

    def to_polygon(self):
        """
        Return a 4-cornered polygon equivalent to this rectangle
        """
        x,y = self.corners.T
        vertices = PixCoord(x=x, y=y)
        return PolygonPixelRegion(vertices=vertices, meta=self.meta,
                                  visual=self.visual)


    def _lower_left_xy(self):
        """
        Compute lower left `xy` position.

        This is used for the conversion to matplotlib in ``as_artist``

        Taken from http://photutils.readthedocs.io/en/latest/_modules/photutils/aperture/rectangle.html#RectangularAperture.plot
        """
        hw = self.width / 2.
        hh = self.height / 2.
        sint = np.sin(self.angle)
        cost = np.cos(self.angle)
        dx = (hh * sint) - (hw * cost)
        dy = -(hh * cost) - (hw * sint)
        x = self.center.x + dx
        y = self.center.y + dy
        return x, y


class RectangleSkyRegion(SkyRegion):
    """
    A rectangle in sky coordinates.

    Parameters
    ----------
    center : `~astropy.coordinates.SkyCoord`
        The position of the center of the rectangle.
    width : `~astropy.units.Quantity`
        The width of the rectangle (before rotation) as an angle
    height : `~astropy.units.Quantity`
        The height of the rectangle (before rotation) as an angle
    angle : `~astropy.units.Quantity`, optional
        The rotation angle of the rectangle, measured anti-clockwise. If set to
        zero (the default), the width axis is lined up with the longitude axis
        of the celestial coordinates.
    meta : `~regions.RegionMeta` object, optional
        A dictionary which stores the meta attributes of this region.
    visual : `~regions.RegionVisual` object, optional
        A dictionary which stores the visual meta attributes of this region.
    """

    center = ScalarSky('center')
    width = QuantityLength('width')
    height = QuantityLength('height')
    angle = QuantityLength('angle')

    def __init__(self, center, width, height, angle=0 * u.deg, meta=None, visual=None):
        self.center = center
        self.width = width
        self.height = height
        self.angle = angle
        self.meta = meta or {}
        self.visual = visual or {}
        self._repr_params = ('width', 'height', 'angle')

    def to_pixel(self, wcs):
        center, scale, north_angle = skycoord_to_pixel_scale_angle(self.center, wcs)
        # FIXME: The following line is needed to get a scalar PixCoord
        center = PixCoord(float(center.x), float(center.y))
        width = self.width.to('deg').value * scale
        height = self.height.to('deg').value * scale
        return RectanglePixelRegion(center, width, height,
                                    angle=self.angle + (north_angle - 90 * u.deg),
                                    meta=self.meta, visual=self.visual)




import distutils.version as vers
import pytest

from numpy.testing import assert_allclose
import astropy.version as astrov
from astropy.utils.data import get_pkg_data_filename
from astropy.table import Table
from astropy.coordinates import SkyCoord
from astropy import units as u
from astropy.io import fits
from astropy.wcs import WCS

from ..read import FITSRegionParser, read_fits_region
from ..write import fits_region_objects_to_table
from ..core import FITSRegionParserError
from ...core import to_shape_list

from ....shapes.circle import CircleSkyRegion

implemented_region_types = ('ellipse', 'circle', 'box', 'polygon', 'point',
                            'annulus', 'elliptannulus')


@pytest.mark.parametrize('filename', ['data/fits_region.fits'])
def test_file_fits(filename):

    filename = get_pkg_data_filename(filename)
    table = Table.read(filename)

    shapes = FITSRegionParser(table, 'warn').shapes

    assert shapes[0].region_type == 'circle'
    assert shapes[1].region_type == 'rectangle'
    assert shapes[2].region_type == 'ellipse'
    assert shapes[3].region_type == 'rectangle'
    assert shapes[4].region_type == 'circleannulus'
    assert shapes[5].region_type == 'point'
    assert shapes[6].region_type == 'point'
    assert shapes[7].region_type == 'polygon'

    for x in range(7):
        assert_allclose(shapes[x].coord[:2],
                        [table['X'][x][0], table['Y'][x][0]])

    assert_allclose(shapes[7].coord, [1, 5, 2, 6, 3, 7, 4, 8])

    assert_allclose(shapes[0].coord[2:], [table['R'][0][0]])

    for x in range(1, 4):
        assert_allclose([val.value for val in shapes[x].coord[2:]],
                        list(table['R'][x][:2]) + [table['ROTANG'][x]])

    regs = shapes.to_regions()
    table_ouput = fits_region_objects_to_table(regs)

    shape_ouput = FITSRegionParser(table_ouput).shapes

    for i in range(len(shapes)):
        assert shapes[i].region_type == shape_ouput[i].region_type
        assert shapes[i].coord == shape_ouput[i].coord
        assert shapes[i].meta == shape_ouput[i].meta

    # Reading the regions directly from file and converting to sky regions.
    regs_sky = read_fits_region(filename)
    header = fits.open(filename)[1].header
    wcs = WCS(header, keysel=['image', 'binary', 'pixel'])
    regs_pix = [reg.to_pixel(wcs) for reg in regs_sky]
    shapes_roundtrip = to_shape_list(regs_pix, 'image')

    for i in range(len(shapes)):
        assert shapes[i].region_type == shapes_roundtrip[i].region_type
        assert_allclose(shapes[i].coord[:-1], shapes_roundtrip[i].coord[:-1])


def test_only_pixel_regions():

    reg_sky = CircleSkyRegion(SkyCoord(1, 2, unit='deg'), 5 * u.deg)

    with pytest.raises(TypeError) as err:
        fits_region_objects_to_table([reg_sky])

    print(str(err))
    assert 'Every region must be a pixel region' in str(err)


def test_valid_columns():

    t = Table([[1, 2, 3]], names=('a'))

    with pytest.raises(FITSRegionParserError) as err:
        FITSRegionParser(t)

    assert "This table has an invalid column name: 'a'" in str(err)


def test_valid_row():

    x = [1]
    y = [2]
    shapes = ['CIRCLE']
    t = Table([x, y, shapes], names=('X', 'Y', 'SHAPE'))
    t['X'].unit = 'pix'
    t['Y'].unit = 'pix'

    with pytest.raises(FITSRegionParserError) as err:
        FITSRegionParser(t)

    assert "The column: 'R' is missing in the table" in str(err)

    t[0]['SHAPE'] = 'PONT'

    with pytest.raises(FITSRegionParserError) as err:
        FITSRegionParser(t)

    assert "'PONT' is not a valid FITS Region type" in str(err)

    t['ROTANG'] = [[20, 30]]
    t['ROTANG'].unit = 'deg'
    t[0]['SHAPE'] = 'PIE'

    with pytest.raises(FITSRegionParserError) as err:
        FITSRegionParser(t)

    assert "'PIE' is currently not supported in regions" in str(err)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import abc
import six
import weakref

from astropy.coordinates import SkyCoord
from astropy.units import Quantity

import numpy as np

from .pixcoord import PixCoord
from .core import PixelRegion, SkyRegion

__all__ = ['RegionMeta', 'RegionVisual']

"""
Several custom descriptor classes are present here for attribute validation of
region classes.
Also, contains RegionMeta and RegionVisual classes to handle meta data of regions.
"""


@six.add_metaclass(abc.ABCMeta)
class RegionAttr(object):
    """
    Meta descriptor class for attribute of an `~regions.Region`
    which makes sure that it's value is valid all the time.
    """

    def __init__(self, name):
        self._name = name

        # WeakKeyDictionary has an object instance as the key
        # and the key value pair remains until the object instance is not
        # free from memory, thus prevents memory leak.
        self._values = weakref.WeakKeyDictionary()

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return self._values.get(instance, None)

    def __set__(self, instance, value):
        self._validate(value)
        self._values[instance] = value

    def _validate(self, value):
        """
        Validates the value of the attribute. Raises an exception if invalid
        else does nothing.
        """
        raise NotImplementedError


class ScalarPix(RegionAttr):
    """
    Descriptor class for `~regions.PixelRegion` which takes a scalar
    `~regions.PixCoord` object.
    """

    def _validate(self, value):
        if not(isinstance(value, PixCoord) and value.isscalar):
            raise ValueError('The {} must be a 0D PixCoord object'
                             .format(self._name))


class OneDPix(RegionAttr):
    """
    Descriptor class for `~regions.PixelRegion` which takes a one dimensional
    `regions.PixCoord` object.
    """

    def _validate(self, value):
        if not(isinstance(value, PixCoord) and not value.isscalar
               and value.x.ndim == 1):
            raise ValueError('The {} must be a 1D PixCoord object'
                             .format(self._name))


class AnnulusCenterPix(object):
    """
    This descriptor class is for the ``center`` of an
    ``annulus`` `~regions.PixelRegion`. It takes a scalar
    `~regions.PixCoord` object. It also makes sure that ``region1`` and
    ``region2`` are in sync in case of an update.
    """

    def __get__(self, instance, owner):
        if instance is None:
            return self
        reg1 = getattr(instance, 'region1')
        return getattr(reg1, 'center')

    def __set__(self, instance, value):

        reg1 = getattr(instance, 'region1')
        reg2 = getattr(instance, 'region2')

        if isinstance(value, PixCoord) and value.isscalar:
            setattr(reg1, 'center', value)
            setattr(reg2, 'center', value)
        else:
            raise ValueError('The center must be a 0D PixCoord object')


class ScalarLength(RegionAttr):
    """
    Descriptor class for `~regions.PixelRegion` which takes a scalar
    python/numpy number.
    """

    def _validate(self, value):
        if not np.isscalar(value):
            raise ValueError(
                'The {} must be a scalar numpy/python number'.format(self._name))


class AnnulusInnerScalarLength(object):
    """
    This descriptor class is for an inner length of an
    ``annulus`` `~regions.PixelRegion`. It takes a scalar
    python/numpy number and makes sure that it is less than the outer
    length of the annulus.
    """

    def __init__(self, name):
        self._name = name

    def __get__(self, instance, owner):
        if instance is None:
            return self
        reg1 = getattr(instance, 'region1')
        return getattr(reg1, self._name)

    def __set__(self, instance, value):
        reg1 = getattr(instance, 'region1')
        reg2 = getattr(instance, 'region2')

        if np.isscalar(value):
            if getattr(reg2, self._name) < value:
                raise ValueError("The inner {0} must be less than the outer {0}"
                                .format(self._name)
                             )
            else:
                setattr(reg1, self._name, value)
        else:
            raise ValueError('The inner {} must be a scalar numpy/python number'
                             .format(self._name))


class AnnulusOuterScalarLength(object):
    """
    This descriptor class is for an outer length of an
    ``annulus`` `~regions.PixelRegion`. It takes a scalar
    python/numpy number and makes sure that it is greater than the inner
    length of the annulus.
    """

    def __init__(self, name):
        self._name = name

    def __get__(self, instance, owner):
        if instance is None:
            return self
        reg2 = getattr(instance, 'region2')
        return getattr(reg2, self._name)

    def __set__(self, instance, value):
        reg1 = getattr(instance, 'region1')
        reg2 = getattr(instance, 'region2')

        if np.isscalar(value):
            if getattr(reg1, self._name) > value:
                raise ValueError("The outer {0} must be greater than the outer"
                                 " {0}".format(self._name)
                                 )
            else:
                setattr(reg2, self._name, value)
        else:
            raise ValueError('The outer {} must be a scalar numpy/python number'
                             .format(self._name))


class ScalarSky(RegionAttr):
    """
    Descriptor class for `~regions.SkyRegion` which takes a scalar
    `~astropy.coordinates.SkyCoord` object.
    """

    def _validate(self, value):
        if not(isinstance(value, SkyCoord) and value.isscalar):
            raise ValueError('The {} must be a 0D SkyCoord object'.
                             format(self._name))


class OneDSky(RegionAttr):
    """
    Descriptor class for `~regions.SkyRegion` which takes a one dimensional
    `~astropy.coordinates.SkyCoord` object.
    """

    def _validate(self, value):
        if not(isinstance(value, SkyCoord) and value.ndim == 1):
            raise ValueError('The {} must be a 1D SkyCoord object'.
                             format(self._name))


class AnnulusCenterSky(object):
    """
    This descriptor class is for the ``center`` of an
    ``annulus`` `~regions.SkyRegion`. It takes a scalar
    `~astropy.coordinates.SkyCoord` object. It also makes sure that ``region1``
    and ``region2`` are in sync in case of an update.
    """

    def __get__(self, instance, owner):
        if instance is None:
            return self
        reg1 = getattr(instance, 'region1')
        return getattr(reg1, 'center')

    def __set__(self, instance, value):

        reg1 = getattr(instance, 'region1')
        reg2 = getattr(instance, 'region2')

        if isinstance(value, SkyCoord) and value.isscalar:
            setattr(reg1, 'center', value)
            setattr(reg2, 'center', value)
        else:
            raise ValueError('The center must be a 0D SkyCoord object')


class QuantityLength(RegionAttr):
    """
    Descriptor class for `~regions.SkyRegion`  which takes a scalar
    `~astropy.units.Quantity` object.
    """

    def _validate(self, value):
        if not(isinstance(value, Quantity) and value.isscalar):
            raise ValueError('The {} must be a scalar astropy Quantity object'
                             .format(self._name))


class AnnulusInnerQuantityLength(object):
    """
    This descriptor class is for an inner length of an ``annulus``
    `~regions.SkyRegion`. It takes a scalar `astropy.units.Quantity` object and
    makes sure that it is less than the outer length of the annulus.
    """

    def __init__(self, name):
        self._name = name

    def __get__(self, instance, owner):
        if instance is None:
            return self
        reg1 = getattr(instance, 'region1')
        return getattr(reg1, self._name)

    def __set__(self, instance, value):
        reg1 = getattr(instance, 'region1')
        reg2 = getattr(instance, 'region2')

        if isinstance(value, Quantity) and value.isscalar:
            if getattr(reg2, self._name) < value:
                raise ValueError("The inner {0} must be less than the outer {0}"
                                 .format(self._name)
                                 )
            else:
                setattr(reg1, self._name, value)
        else:
            raise ValueError('The inner {} must be a scalar astropy Quantity '
                             'object'.format(self._name))


class AnnulusOuterQuantityLength(object):
    """
    This descriptor class is for an outer length of an ``annulus``
    `~regions.SkyRegion`. It takes a scalar `astropy.units.Quantity` object
    and makes sure that it is less than the inner length of the annulus.
    """

    def __init__(self, name):
        self._name = name

    def __get__(self, instance, owner):
        if instance is None:
            return self
        reg2 = getattr(instance, 'region2')
        return getattr(reg2, self._name)

    def __set__(self, instance, value):
        reg1 = getattr(instance, 'region1')
        reg2 = getattr(instance, 'region2')

        if isinstance(value, Quantity) and value.isscalar:
            if getattr(reg1, self._name) > value:
                raise ValueError("The inner {0} must be less than the outer {0}"
                                 .format(self._name)
                                 )
            else:
                setattr(reg2, self._name, value)
        else:
            raise ValueError('The outer {} must be a scalar astropy Quantity '
                             'object'.format(self._name))


class AnnulusAngle(object):
    """
    This descriptor class is for the ``center`` of an ``annulus``
    `~regions.SkyRegion`. It takes a scalar `~astropy.units.Quantity` object.
    It also makes sure that ``region1`` and ``region2`` are in sync in case of
    an update.
    """

    def __get__(self, instance, owner):
        if instance is None:
            return self
        reg1 = getattr(instance, 'region1')
        return getattr(reg1, 'angle')

    def __set__(self, instance, value):

        reg1 = getattr(instance, 'region1')
        reg2 = getattr(instance, 'region2')

        if isinstance(value, Quantity) and value.isscalar:
            setattr(reg1, 'angle', value)
            setattr(reg2, 'angle', value)
        else:
            raise ValueError('The angle must be a scalar astropy quantity object')


class CompoundRegionPix(RegionAttr):
    """
    Descriptor class for `~regions.CompoundPixelRegion` which takes a
    `~regions.PixelRegion` object.
    """

    def _validate(self, value):
        if not isinstance(value, PixelRegion):
            raise ValueError('The {} must be a PixelRegion object'
                             .format(self._name))


class CompoundRegionSky(RegionAttr):
    """
    Descriptor class is for `~regions.CompoundSkyRegion` which takes a
    `~regions.SkyRegion` object.
    """

    def _validate(self, value):
        if not isinstance(value, SkyRegion):
            raise ValueError('The {} must be a SkyRegion object'
                             .format(self._name))


@six.add_metaclass(abc.ABCMeta)
class Meta(dict):

    def __init__(self, seq=None, **kwargs):

        super(Meta, self).__init__()

        if seq:
            if isinstance(seq, dict):
                for key, val in seq.items():
                    self.__setitem__(key, val)
            else:
                for key, val in seq:
                    self.__setitem__(key, val)

        if len(kwargs) > 0:
            for key, val in kwargs.items():
                self.__setitem__(key, val)

    def __setitem__(self, key, value):
        key = self.key_mapping.get(key, key)
        if key in self.valid_keys:
            super(Meta, self).__setitem__(key, value)
        else:
            raise KeyError(
                "{} is not a valid key for this class.".format(key))

    def __getitem__(self, item):
        item = self.key_mapping.get(item, item)
        return super(Meta, self).__getitem__(item)


class RegionMeta(Meta):
    """
    A python dictionary subclass which holds the meta attributes of the region.
    """
    valid_keys = ['label', 'include', 'frame', 'range', 'veltype',
                  'restfreq', 'tag', 'comment', 'line', 'name',
                  'select', 'highlite', 'fixed', 'edit', 'move', 'rotate',
                  'delete', 'source', 'background', 'corr', 'type'
                  ]

    key_mapping = {}


class RegionVisual(Meta):
    """
    A python dictionary subclass which holds the visual attributes of the region.
    """
    valid_keys = ['color', 'dash', 'font', 'dashlist', 'symsize', 'symthick',
                  'symbol', 'fontsize', 'fontstyle', 'usetex', 'labelpos',
                  'labeloff', 'linewidth', 'linestyle', 'fill', 'line',
                  'textangle', 'fontweight']

    key_mapping = {'width': 'linewidth', 'point': 'symbol'}



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Geometry subpackage for low-level geometry functions.
"""

from .rotate_polygon import *
from .circular_overlap import *
from .elliptical_overlap import *
from .rectangular_overlap import *
from .polygonal_overlap import *


__all__ = ['circular_overlap_grid', 'elliptical_overlap_grid',
           'rectangular_overlap_grid', 'polygonal_overlap_grid']



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import distutils.version as vers

from numpy.testing import assert_allclose
import pytest

from astropy.utils.data import get_pkg_data_filename, get_pkg_data_filenames
import astropy.version as astrov
from astropy.coordinates import Angle, SkyCoord
from astropy.tests.helper import catch_warnings, assert_quantity_allclose
from astropy.utils.exceptions import AstropyUserWarning
from astropy import units as u

from ....shapes.circle import CircleSkyRegion
from ..read import read_ds9, DS9Parser
from ..write import write_ds9, ds9_objects_to_string

_ASTROPY_MINVERSION = vers.LooseVersion('1.1')
_ASTROPY_VERSION = vers.LooseVersion(astrov.version)


@pytest.mark.xfail(_ASTROPY_VERSION < _ASTROPY_MINVERSION,
                   reason='Some coordinates systems not available in older version of astropy')
def test_read():
    # Check that all test files including reference files are readable
    files = get_pkg_data_filenames('data')
    for f in files:
        with open(f) as f:
            DS9Parser(f.read(), errors='warn')


implemented_region_types = ('ellipse', 'circle', 'rectangle', 'polygon', 'point')


@pytest.mark.parametrize('filename',
                         ['data/ds9.fk5.reg',
                          'data/ds9.fk5.hms.reg',
                          'data/ds9.fk5.hms.strip.reg',
                          'data/ds9.fk5.strip.reg',
                          'data/ds9.galactic.reg',
                          'data/ds9.galactic.hms.reg',
                          'data/ds9.galactic.hms.strip.reg',
                          'data/ds9.galactic.strip.reg',
                          # TODO : data/ds9.physical.windows.reg contains different values -> Why?
                          'data/ds9.physical.reg',
                          'data/ds9.physical.strip.reg',
                         ])
def test_file(filename):
    filename = get_pkg_data_filename(filename)
    regs = read_ds9(filename, errors='warn')

    coordsys = os.path.basename(filename).split(".")[1]

    actual = ds9_objects_to_string(regs, coordsys=str(coordsys), fmt='.2f',
                                   radunit=None if coordsys=='physical' else 'arcsec').strip()

    reffile = get_pkg_data_filename('data/{coordsys}{strip}_reference.reg'
                                    .format(coordsys=coordsys,
                                            strip=("_strip"
                                                   if "strip" in filename
                                                   else "")))

    with open(reffile, 'r') as fh:
        desired = fh.read().strip()

    # since metadata is not required to preserve order, we have to do a more
    # complex comparison
    desired_lines = [set(line.split(" ")) for line in desired.split("\n")]
    actual_lines = [set(line.split(" ")) for line in actual.split("\n")]
    for split_line in actual_lines:
        assert split_line in desired_lines

    for split_line in desired_lines:
        assert split_line in actual_lines


def test_ds9_objects_to_str():
    """
    Simple test case for ds9_objects_to_str.
    """
    center = SkyCoord(42, 43, unit='deg')
    radius = Angle(3, 'deg')
    region = CircleSkyRegion(center, radius)
    expected = '# Region file format: DS9 astropy/regions\nfk5\ncircle(42.0000,43.0000,3.0000)\n'
    actual = ds9_objects_to_string([region], fmt='.4f')
    assert actual == expected


def test_ds9_string_to_objects():
    """
    Simple test case for ds9_string_to_objects.
    """
    ds9_str = '# Region file format: DS9 astropy/regions\nfk5\ncircle(42.0000,43.0000,3.0000)\n'
    parser = DS9Parser(ds9_str)
    regions = parser.shapes.to_regions()
    reg = regions[0]

    assert_allclose(reg.center.ra.deg, 42)
    assert_allclose(reg.center.dec.deg, 43)
    assert_allclose(reg.radius.value, 3)


def test_ds9_string_to_objects_whitespace():
    """
    Simple test case for ds9_string_to_objects.
    """
    ds9_str = '# Region file format: DS9 astropy/regions\nfk5\n -circle(42.0000,43.0000,3.0000)\n'
    parser = DS9Parser(ds9_str)
    assert parser.shapes[0].include == False


def test_ds9_io(tmpdir):
    """
    Simple test case for write_ds9 and read_ds9.
    """
    center = SkyCoord(42, 43, unit='deg', frame='fk5')
    radius = Angle(3, 'deg')
    reg = CircleSkyRegion(center, radius)
    reg.meta['name'] = 'MyName'

    filename = os.path.join(str(tmpdir), 'ds9.reg')
    write_ds9([reg], filename, coordsys='fk5')
    reg = read_ds9(filename)[0]

    assert_allclose(reg.center.ra.deg, 42)
    assert_allclose(reg.center.dec.deg, 43)
    assert_allclose(reg.radius.value, 3)
    assert 'name' in reg.meta
    assert reg.meta['name'] == 'MyName'


def test_missing_region_warns():
    ds9_str = '# Region file format: DS9 astropy/regions\nfk5\ncircle(42.0000,43.0000,3.0000)\nnotaregiontype(blah)'

    # this will warn on both the commented first line and the not_a_region line
    with catch_warnings(AstropyUserWarning) as ASWarn:
        parser = DS9Parser(ds9_str, errors='warn')

    assert len(parser.shapes) == 1
    assert len(ASWarn) == 1
    assert "Region type 'notaregiontype' was identified, but it is not one of the known region types." in str(ASWarn[0].message)


def test_global_parser():
    """
    Check that the global_parser does what's expected
    """

    # have to force "str" here because unicode_literals makes this string a
    # unicode string, even though in python2.7 we don't want it to be.  But in
    # py3, we don't want it to be bytes.
    global_test_str = str('global color=green dashlist=8 3 width=1'
                          ' font="helvetica 10 normal roman" select=1'
                          ' highlite=1 dash=0 fixed=0 edit=1 move=1'
                          ' delete=1 include=1 source=1')
    global_parser = DS9Parser(global_test_str)
    assert dict(global_parser.global_meta) == {'dash': '0', 'source': '1', 'move': '1',
                                  'font': 'helvetica 10 normal roman',
                                  'dashlist': '8 3', 'include': True,
                                  'highlite': '1', 'color': 'green',
                                  'select': '1',
                                  'fixed': '0', 'width': '1', 'edit': '1',
                                  'delete': '1'}


def test_ds9_color():
    """
    Color parsing test
    """
    ds9_str = '# Region file format: DS9 astropy/regions\nfk5\ncircle(42.0000,43.0000,3.0000) # color=green' \
              '\ncircle(43.0000,43.0000,3.0000) # color=orange\n'

    parser = DS9Parser(ds9_str)
    regions = parser.shapes

    assert regions[0].meta['color'] == 'green'
    assert regions[1].meta['color'] == 'orange'


def test_ds9_color_override_global():
    """
    Color parsing test in the presence of a global
    """
    global_test_str = str('global color=blue dashlist=8 3 width=1'
                          ' font="helvetica 10 normal roman" select=1'
                          ' highlite=1 dash=0 fixed=0 edit=1 move=1'
                          ' delete=1 include=1 source=1')

    ds9_str = '# Region file format: DS9 astropy/regions\n{global_str}\nfk5\n'
    reg1str = "circle(42.0000,43.0000,3.0000) # color=green"
    reg2str = "circle(43.0000,43.0000,3.0000) # color=orange"
    reg3str = "circle(43.0000,43.0000,3.0000)"
    ds9_str = ds9_str.format(global_str=global_test_str) + "\n".join([reg1str, reg2str, reg3str])
    parser = DS9Parser(ds9_str)
    regions = parser.shapes
    assert regions[0].meta['color'] == 'green'
    assert regions[1].meta['color'] == 'orange'
    assert regions[2].meta['color'] == 'blue'


def test_issue134_regression():
    regstr = 'galactic; circle(+0:14:26.064,+0:00:45.206,30.400")'
    parser = DS9Parser(regstr)
    regions = parser.shapes
    assert regions[0].to_region().radius.value == 30.4


def test_issue65_regression():
    regstr = 'J2000; circle 188.5557102 12.0314056 1" # color=red'
    parser = DS9Parser(regstr)
    regions = parser.shapes
    reg = regions[0].to_region()
    assert reg.center.ra.value == 188.5557102
    assert reg.center.dec.value == 12.0314056
    assert reg.radius.value == 1.0


def test_pixel_angle():
    """
    Checks whether angle in PixelRegions is a u.Quantity object.
    """
    reg_str = 'image\nbox(1.5,2,2,1,0)'

    reg = DS9Parser(reg_str).shapes.to_regions()[0]

    assert isinstance(reg.angle, u.quantity.Quantity)


def test_expicit_formatting_directives():
    """
    Checks whether every explicit formatting directive is supported
    """

    reg_str = 'image\ncircle(1.5, 2, 2)'
    valid_coord = DS9Parser(reg_str).shapes[0].coord

    reg_str_explicit = 'image\ncircle(1.5p, 2p, 2p)'
    coord = DS9Parser(reg_str_explicit).shapes[0].coord
    assert_quantity_allclose(coord, valid_coord)

    reg_str_explicit = 'image\ncircle(1.5i, 2i, 2i)'
    coord = DS9Parser(reg_str_explicit).shapes[0].coord
    assert_quantity_allclose(coord, valid_coord)

    reg_str = 'fk5\ncircle(1.5, 2, 2)'
    valid_coord = DS9Parser(reg_str).shapes[0].coord

    reg_str_explicit = 'fk5\ncircle(1.5d, 2d, 2d)'
    coord = DS9Parser(reg_str_explicit).shapes[0].coord
    assert_quantity_allclose(coord, valid_coord)

    reg_str_explicit = 'fk5\ncircle(1.5r, 2r, 2r)'
    coord = DS9Parser(reg_str_explicit).shapes[0].coord
    for val in coord:
        assert val.unit == u.Unit('rad')

    reg_str = 'fk5\ncircle(1:20:30, 2:3:7, 2)'
    valid_coord = DS9Parser(reg_str).shapes[0].coord

    reg_str_explicit = 'fk5\ncircle(1h20m30s, 2d3m7s, 2d)'
    coord = DS9Parser(reg_str_explicit).shapes[0].coord
    assert_quantity_allclose(u.Quantity(coord), u.Quantity(valid_coord))



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import absolute_import, division, print_function
from warnings import warn
import string
import numbers
import numpy as np

from astropy import units as u
from astropy import coordinates
from astropy.coordinates import Angle, SkyCoord
from astropy import log
from astropy.table import Table

from .. import shapes
from ..core import PixCoord, SkyRegion
from ..core.attributes import RegionMeta, RegionVisual
from .ds9.core import DS9RegionParserWarning, valid_symbols_ds9
from .crtf.core import CRTFRegionParserWarning

__all__ = ['ShapeList', 'Shape', 'to_shape_list', 'to_crtf_meta', 'to_ds9_meta']


class RegionConversionError(ValueError):
    """
    A generic error class for Shape to Region conversion.
    """


regions_attributes = dict(circle=['center', 'radius'],
                          ellipse=['center', 'width', 'height', 'angle'],
                          rectangle=['center', 'width', 'height', 'angle'],
                          polygon=['vertices'],
                          circleannulus=['center', 'inner_radius', 'outer_radius'],
                          ellipseannulus=['center', 'inner_width',
                                          'inner_height', 'outer_width',
                                          'outer_height', 'angle'],
                          line=['start', 'end'],
                          point=['center'],
                          text=['center']
                          )
regions_attributes['rectangleannulus'] = regions_attributes['ellipseannulus']

# This helps to map the region names in the respective format to the ones
# available in this package
reg_mapping = {'DS9': {x: x for x in regions_attributes},
               'CRTF': {x: x for x in regions_attributes},
               'FITS_REGION': {x: x for x in regions_attributes}}
reg_mapping['DS9']['box'] = 'rectangle'
reg_mapping['CRTF']['rotbox'] = 'rectangle'
reg_mapping['CRTF']['box'] = 'rectangle'
reg_mapping['CRTF']['centerbox'] = 'rectangle'
reg_mapping['CRTF']['poly'] = 'polygon'
reg_mapping['CRTF']['symbol'] = 'point'
reg_mapping['CRTF']['text'] = 'text'
reg_mapping['CRTF']['annulus'] = 'circleannulus'
reg_mapping['DS9']['text'] = 'text'
reg_mapping['DS9']['annulus'] = 'circleannulus'
reg_mapping['FITS_REGION']['annulus'] = 'circleannulus'
reg_mapping['FITS_REGION']['box'] = 'rectangle'
reg_mapping['FITS_REGION']['rotbox'] = 'rectangle'
reg_mapping['FITS_REGION']['elliptannulus'] = 'ellipseannulus'

# valid astropy coordinate frames in their respective formats.
valid_coordsys = {'DS9': ['image', 'physical', 'fk4', 'fk5', 'icrs', 'galactic',
                          'geocentrictrueecliptic', 'wcs'],
                  'CRTF': ['image', 'fk5', 'fk4', 'galactic',
                           'geocentrictrueecliptic', 'supergalactic', 'icrs']
                  }
valid_coordsys['DS9'] += ['wcs{}'.format(x) for x in string.ascii_lowercase]

# Maps astropy's coordinate frame names with their respective name in the file format.
coordsys_mapping = {'DS9': {x: x for x in valid_coordsys['DS9']},
                    'CRTF': {x: x for x in valid_coordsys['CRTF']}
                    }
coordsys_mapping['CRTF']['geocentrictrueecliptic'] = 'ecliptic'
coordsys_mapping['CRTF']['fk5'] = 'j2000'
coordsys_mapping['CRTF']['fk4'] = 'b1950'
coordsys_mapping['CRTF']['supergalactic'] = 'supergal'

coordsys_mapping['DS9']['geocentrictrueecliptic'] = 'ecliptic'


class ShapeList(list):
    """List of Shape"""
    def to_regions(self):
        regions = list()
        for shape in self:
            # Skip elliptical annulus for now
            if shape.region_type == 'ellipse' and len(shape.coord) > 5:
                msg = 'Skipping elliptical annulus {}'.format(shape)
                warn(msg, DS9RegionParserWarning)
                continue
            if shape.region_type in ['box'] and shape.format_type == 'CRTF':
                msg = 'Skipping {} {}'.format(shape.region_type, shape)
                warn(msg, CRTFRegionParserWarning)
                continue
            log.debug(shape)
            region = shape.to_region()
            log.debug(region)
            regions.append(region)
        return regions

    def to_crtf(self,  coordsys='fk5', fmt='.6f', radunit='deg'):
        """
        Converts a list of ``regions.Shape`` objects to crtf region strings.

        Parameters
        ----------
        coordsys : str
            This overrides the coordinate system frame for all regions.

        fmt : str
            A python string format defining the output precision.
            Default is .6f, which is accurate to 0.0036 arcseconds.

        radunit : str
            This denotes the unit of the radius.

        Returns
        -------
        region_string : str
            crtf region string

        Examples
        --------
        TODO
        """

        crtf_strings = {
            'circle': '{0}circle[[{1:FMT}deg, {2:FMT}deg], {3:FMT}RAD]',
            'circleannulus': '{0}annulus[[{1:FMT}deg, {2:FMT}deg], [{3:FMT}RAD, {4:FMT}RAD]]',
            'ellipse': '{0}ellipse[[{1:FMT}deg, {2:FMT}deg], [{3:FMT}RAD, {4:FMT}RAD], {5:FMT}deg]',
            'rectangle': '{0}rotbox[[{1:FMT}deg, {2:FMT}deg], [{3:FMT}RAD, {4:FMT}RAD], {5:FMT}deg]',
            'polygon': '{0}poly[{1}]',
            'point': '{0}point[[{1:FMT}deg, {2:FMT}deg]]',
            'symbol': '{0}symbol[[{1:FMT}deg, {2:FMT}deg], {symbol}]',
            'text': '{0}text[[{1:FMT}deg, {2:FMT}deg], \'{text}\']',
            'line': '{0}line[[{1:FMT}deg, {2:FMT}deg], [{3:FMT}deg, {4:FMT}deg]]'
                        }

        output = '#CRTF\n'

        if radunit == 'arcsec':
            # what's this for?
            if coordsys in coordsys_mapping['CRTF'].values():
                radunitstr = '"'
            else:
                raise ValueError(
                    'Radius unit arcsec not valid for coordsys {}'.format(
                        coordsys))
        else:
            radunitstr = radunit

        for key, val in crtf_strings.items():
            crtf_strings[key] = val.replace("FMT", fmt).replace("RAD",
                                                               radunitstr)

        output += 'global coord={}\n'.format(coordsys)

        for shape in self:

            shape.check_crtf()
            shape.meta = to_crtf_meta(shape.meta)

            # if unspecified, include is True.
            include = "-" if shape.include in (False, '-') else "+"
            include += "ann " if shape.meta.get('type', 'reg') == 'ann' else ""

            if shape.meta.get('label', "") != "":
                shape.meta['label'] = "'{}'".format(shape.meta['label'])
            meta_str = ", ".join("{0}={1}".format(key, val) for key, val in
                                shape.meta.items() if
                                key not in ('include', 'comment', 'symbol',
                                            'coord', 'text', 'range', 'corr',
                                            'type', 'text'))

            if 'comment' in shape.meta:
                meta_str += ", " + shape.meta['comment']
            if 'range' in shape.meta:
                shape.meta['range'] = [str(str(x).replace(" ", "")) for x in
                                       shape.meta['range']]
                meta_str += ", range={}".format(shape.meta['range']).replace("'", "")
            if 'corr' in shape.meta:
                meta_str += ", corr={}".format(shape.meta['corr']).replace("'", "")

            coord = []
            if coordsys not in ['image', 'physical']:
                for val in shape.coord:
                    if isinstance(val, Angle):
                        coord.append(float(val.value))
                    else:
                        if radunit == '' or None:
                            coord.append(float(val.value))
                        else:
                            coord.append(float(val.to(radunit).value))
            else:
                for val in shape.coord:
                    if isinstance(val, u.Quantity):
                        coord.append(float(val.value))
                    else:
                        coord.append(float(val))

            if shape.region_type in ['ellipse', 'rectangle'] and len(shape.coord) % 2 == 1:
                coord[-1] = float(shape.coord[-1].to('deg').value)

            if shape.region_type == 'polygon':
                val = '[{0:' + fmt + '}deg, {1:' + fmt + '}deg]'
                temp = [val.format(x, y) for x, y in zip(coord[::2], coord[1::2])]
                coord = ", ".join(temp)
                line = crtf_strings['polygon'].format(include, coord)
            elif shape.region_type == 'point':
                if 'symbol' in shape.meta:
                    line = crtf_strings['symbol'].format(include, *coord,
                                                         symbol=shape.meta['symbol'])
                else:
                    line = crtf_strings['point'].format(include, *coord)
            elif shape.region_type == 'ellipse':
                coord[2:] = [x / 2 for x in coord[2:]]
                if len(coord) % 2 == 1:
                    coord[-1] *= 2
                line = crtf_strings['ellipse'].format(include, *coord)
            elif shape.region_type == 'text':
                line = crtf_strings['text'].format(include, *coord, text=shape.meta['text'])
            else:
                line = crtf_strings[shape.region_type].format(include, *coord)

            if meta_str.strip():
                output += "{0}, {1}\n".format(line, meta_str)
            else:
                output += "{0}\n".format(line)

        return output

    def to_ds9(self, coordsys='fk5', fmt='.6f', radunit='deg'):
        """
        Converts a list of ``regions.Shape`` objects to ds9 region strings.

        Parameters
        ----------
        coordsys : str
            This overrides the coordinate system frame for all regions.
        fmt : str
            A python string format defining the output precision.
            Default is .6f, which is accurate to 0.0036 arcseconds.
        radunit : str
            This denotes the unit of the radius.

        Returns
        -------
        region_string : str
            ds9 region string

        Examples
        --------
        TODO
        """
        valid_symbols_reverse = {y: x for x, y in valid_symbols_ds9.items()}

        ds9_strings = {
            'circle': '{0}circle({1:FMT},{2:FMT},{3:FMT}RAD)',
            'circleannulus': '{0}annulus({1:FMT},{2:FMT},{3:FMT}RAD,{4:FMT}RAD)',
            'ellipse': '{0}ellipse({1:FMT},{2:FMT},{3:FMT}RAD,{4:FMT}RAD,{5:FMT})',
            'rectangle': '{0}box({1:FMT},{2:FMT},{3:FMT}RAD,{4:FMT}RAD,{5:FMT})',
            'polygon': '{0}polygon({1})',
            'point': '{0}point({1:FMT},{2:FMT})',
            'line': '{0}line({1:FMT},{2:FMT},{3:FMT},{4:FMT})',
            'text': '{0}text({1:FMT},{2:FMT})'
                      }

        output = '# Region file format: DS9 astropy/regions\n'

        if radunit == 'arcsec':
            # what's this for?
            if coordsys in coordsys_mapping['DS9'].values():
                radunitstr = '"'
            else:
                raise ValueError('Radius unit arcsec not valid for coordsys {}'.format(coordsys))
        else:
            radunitstr = ''

        for key, val in ds9_strings.items():
            ds9_strings[key] = val.replace("FMT", fmt).replace("RAD", radunitstr)

        output += '{}\n'.format(coordsys)

        for shape in self:

            shape.check_ds9()
            shape.meta = to_ds9_meta(shape.meta)

            # if unspecified, include is True.
            include = "-" if shape.include in (False, '-') else ""

            if 'point' in shape.meta:
                shape.meta['point'] = valid_symbols_reverse[shape.meta['point']]

            if 'symsize' in shape.meta:
                shape.meta['point'] += " {}".format(shape.meta.pop('symsize'))

            meta_str = " ".join("{0}={1}".format(key, val) for key, val in
                                shape.meta.items() if key not in ('include', 'tag', 'comment', 'font', 'text'))
            if 'tag' in shape.meta:
                meta_str += " " + " ".join(["tag={0}".format(tag) for tag in shape.meta['tag']])
            if 'font' in shape.meta:
                meta_str += " " + 'font="{0}"'.format(shape.meta['font'])
            if shape.meta.get('text', '') != '':
                meta_str += " " + 'text={' + shape.meta['text'] + '}'
            if 'comment' in shape.meta:
                meta_str += " " + shape.meta['comment']

            coord = []

            if coordsys not in ['image', 'physical']:
                for val in shape.coord:
                    if isinstance(val, Angle):
                        coord.append(float(val.value))
                    else:
                        if radunit == '' or None:
                            coord.append(float(val.value))
                        else:
                            coord.append(float(val.to(radunit).value))
                if shape.region_type in ['ellipse', 'rectangle'] and len(shape.coord) % 2 == 1:
                    coord[-1] = float(shape.coord[-1].to('deg').value)
            else:
                for val in shape.coord:
                    if isinstance(val, u.Quantity):
                        coord.append(float(val.value))
                    else:
                        coord.append(float(val))
                if shape.region_type in ['polygon', 'line']:
                    coord = [x+1 for x in coord]
                else:
                    coord[0] += 1
                    coord[1] += 1

            if shape.region_type == 'polygon':
                val = "{0:" + fmt + "}"
                temp = [val.format(x) for x in coord]
                coord = ",".join(temp)
                line = ds9_strings['polygon'].format(include, coord)
            elif shape.region_type == 'ellipse':
                coord[2:] = [x / 2 for x in coord[2:]]
                if len(coord) % 2 == 1:
                    coord[-1] *= 2
                line = ds9_strings['ellipse'].format(include, *coord)
            else:
                line = ds9_strings[shape.region_type].format(include, *coord)

            if meta_str.strip():
                output += "{0} # {1}\n".format(line, meta_str)
            else:
                output += "{0}\n".format(line)

        return output

    def to_fits(self):
        """
        Converts a `~regions.ShapeList` to a `~astropy.table.Table` object.
        """

        max_length_coord = 1
        coord_x = []
        coord_y = []
        shapes = []
        radius = []
        rotangle_deg = []
        components = []

        reg_reverse_mapping = {value: key for key, value in
                               reg_mapping['FITS_REGION'].items()}
        reg_reverse_mapping['rectangle'] = 'ROTBOX'
        reg_reverse_mapping['circleannulus'] = 'ANNULUS'
        reg_reverse_mapping['ellipseannulus'] = 'ELLIPTANNULUS'

        for num, shape in enumerate(self):
            shapes.append(reg_reverse_mapping[shape.region_type])
            if shape.region_type == 'polygon':
                max_length_coord = max(len(shape.coord)/2, max_length_coord)
                coord = [x.value for x in shape.coord]
                coord_x.append(coord[::2])
                coord_y.append(coord[1::2])
                radius.append(0)
                rotangle_deg.append(0)
            else:
                coord_x.append(shape.coord[0].value)
                coord_y.append(shape.coord[1].value)
                if shape.region_type in ['circle', 'circleannulus', 'point']:
                    radius.append([float(val) for val in shape.coord[2:]])
                    rotangle_deg.append(0)
                else:
                    radius.append([float(x) for x in shape.coord[2:-1]])
                    rotangle_deg.append(shape.coord[-1].to('deg').value)

            tag = shape.meta.get('tag', '')
            if tag.isdigit():
                components.append(int(tag))
            else:
                components.append(num + 1)

        # padding every value with zeros at the end to make sure that all values
        # in the column have same length.
        for i in range(len(self)):
            if np.isscalar(coord_x[i]):
                coord_x[i] = np.array([coord_x[i]])
            if np.isscalar(coord_y[i]):
                coord_y[i] = np.array([coord_y[i]])
            if np.isscalar(radius[i]):
                radius[i] = np.array([radius[i]])

            coord_x[i] = np.pad(coord_x[i], (0, int(max_length_coord - len(coord_x[i]))),
                                'constant', constant_values=(0, 0))
            coord_y[i] = np.pad(coord_y[i], (0, int(max_length_coord - len(coord_y[i]))),
                                'constant', constant_values=(0, 0))
            radius[i] = np.pad(radius[i], (0, 4 - len(radius[i])), 'constant',
                               constant_values=(0, 0))

        table = Table([coord_x, coord_y, shapes, radius, rotangle_deg, components],
                       names=('X', 'Y', 'SHAPE', 'R', 'ROTANG', 'COMPONENT'))
        table['X'].unit = 'pix'
        table['Y'].unit = 'pix'
        table['R'].unit = 'pix'
        table['ROTANG'].unit = 'deg'

        return table


class Shape(object):
    """
    Helper class to represent a DS9/CRTF Region.

    This serves as intermediate step in the parsing process.

    Parameters
    ----------
    coordsys : str
        Astropy Coordinate system frame used in the region.
    region_type : str
        Type of the region (as defined in this package).
    coord : list of `~astropy.coordinates.Angle` or `~astropy.units.Quantity`
        Coordinates
    meta : dict
        Meta attributes
    composite : bool
        Composite region
    include : bool
        Include/exclude region
    """

    shape_to_sky_region = dict(circle=shapes.CircleSkyRegion,
                               ellipse=shapes.EllipseSkyRegion,
                               rectangle=shapes.RectangleSkyRegion,
                               polygon=shapes.PolygonSkyRegion,
                               circleannulus=shapes.CircleAnnulusSkyRegion,
                               ellipseannulus=shapes.EllipseAnnulusSkyRegion,
                               rectangleannulus=shapes.RectangleAnnulusSkyRegion,
                               line=shapes.LineSkyRegion,
                               point=shapes.PointSkyRegion,
                               text=shapes.TextSkyRegion
                               )

    shape_to_pixel_region = dict(circle=shapes.CirclePixelRegion,
                                 ellipse=shapes.EllipsePixelRegion,
                                 rectangle=shapes.RectanglePixelRegion,
                                 polygon=shapes.PolygonPixelRegion,
                                 circleannulus=shapes.CircleAnnulusPixelRegion,
                                 ellipseannulus=shapes.EllipseAnnulusPixelRegion,
                                 rectangleannulus=shapes.RectangleAnnulusPixelRegion,
                                 line=shapes.LinePixelRegion,
                                 point=shapes.PointPixelRegion,
                                 text=shapes.TextPixelRegion
                                 )

    error = RegionConversionError

    def __init__(self, coordsys, region_type, coord, meta, composite, include):
        self._coordsys = coordsys
        self._region_type = region_type
        self.coord = coord
        self.meta = meta
        self.composite = composite
        self.include = include

    @property
    def coordsys(self):
        return self._coordsys

    @coordsys.setter
    def coordsys(self, value):
        self._coordsys = value.lower()
        self._validate()

    @property
    def region_type(self):
        return self._region_type

    @region_type.setter
    def region_type(self, value):
        self._region_type = value.lower()
        self._validate()

    def __str__(self):
        ss = self.__class__.__name__
        ss += '\nType : {}'.format(self.meta.get('type', 'reg'))
        ss += '\nCoord sys : {}'.format(self.coordsys)
        ss += '\nRegion type : {}'.format(self.region_type)
        if self.region_type == 'symbol':
            ss += '\nSymbol : {}'.format(self.meta['symbol'])
        if self.region_type == 'text':
            ss += '\nText : {}'.format(self.meta['text'])
        ss += '\nMeta: {}'.format(self.meta)
        ss += '\nComposite: {}'.format(self.composite)
        ss += '\nInclude: {}'.format(self.include)
        ss += '\n'
        return ss

    def convert_coords(self):
        """
        Process list of coordinates

        This mainly searches for tuple of coordinates in the coordinate list and
        creates a SkyCoord or PixCoord object from them if appropriate for a
        given region type. This involves again some coordinate transformation,
        so this step could be moved to the parsing process
        """
        if self.coordsys in ['image', 'physical']:
            coords = self._convert_pix_coords()
        else:
            coords = self._convert_sky_coords()

        if self.region_type == 'line':
            coords = [coords[0][0], coords[0][1]]

        if self.region_type == 'text':
            coords.append(self.meta['text'])

        return coords

    def _convert_sky_coords(self):
        """
        Convert to sky coordinates
        """
        parsed_angles = [(x, y)
                         for x, y in zip(self.coord[:-1:2], self.coord[1::2])
                         if (isinstance(x, coordinates.Angle) and isinstance(y, coordinates.Angle))
                         ]
        frame = coordinates.frame_transform_graph.lookup_name(self.coordsys)

        lon, lat = zip(*parsed_angles)
        if hasattr(lon, '__len__') and hasattr(lat, '__len__') and len(lon) == 1 and len(lat) == 1:
            # force entries to be scalar if they are length-1
            lon, lat = u.Quantity(lon[0]), u.Quantity(lat[0])
        else:
            # otherwise, they are vector quantities
            lon, lat = u.Quantity(lon), u.Quantity(lat)
        sphcoords = coordinates.UnitSphericalRepresentation(lon, lat)
        coords = [SkyCoord(frame(sphcoords))]

        if self.region_type != 'polygon':
            coords += self.coord[len(coords * 2):]

        return coords

    def _convert_pix_coords(self):
        """
        Convert to pixel coordinates, `regions.PixCoord`
        """
        if self.region_type in ['polygon', 'line']:
            # have to special-case polygon in the phys coord case
            # b/c can't typecheck when iterating as in sky coord case
            coords = [PixCoord(self.coord[0::2], self.coord[1::2])]
        else:
            temp = [_.value for _ in self.coord]
            coord = PixCoord(temp[0], temp[1])
            coords = [coord] + temp[2:]

        # The angle remains as a quantity object.
        # Modulus check makes sure that it works for ellipse/rectangle annulus
        if self.region_type in ['ellipse', 'rectangle'] and len(coords) % 2 == 0:
            coords[-1] = self.coord[-1]

        return coords

    def to_region(self):
        """
        Converts to region, ``regions.Region`` object
        """

        coords = self.convert_coords()
        log.debug(coords)
        viz_keywords = ['color', 'dash', 'dashlist', 'width', 'font', 'symsize',
                        'symbol', 'symsize', 'fontsize', 'fontstyle', 'usetex',
                        'labelpos', 'labeloff', 'linewidth', 'linestyle',
                        'point', 'textangle', 'fontweight']

        if isinstance(coords[0], SkyCoord):
            reg = self.shape_to_sky_region[self.region_type](*coords)
        elif isinstance(coords[0], PixCoord):
            reg = self.shape_to_pixel_region[self.region_type](*coords)
        else:
            self._raise_error("No central coordinate")

        reg.visual = RegionVisual()
        reg.meta = RegionMeta()

        label = self.meta.pop('text', self.meta.get('label', ""))
        if label != '':
            reg.meta['label'] = label
        for key in self.meta:
            if key in viz_keywords:
                reg.visual[key] = self.meta[key]
            else:
                reg.meta[key] = self.meta[key]
        reg.meta['include'] = self.include
        return reg

    def _raise_error(self, msg):
        raise self.error(msg)

    def check_crtf(self):
        """
        Checks for CRTF compatibility.
        """
        if self.region_type not in regions_attributes:
            raise ValueError("'{0}' is not a valid region type in this package"
                             "supported by CRTF".format(self.region_type))

        if self.coordsys not in valid_coordsys['CRTF']:
            raise ValueError("'{0}' is not a valid coordinate reference frame in "
                             "astropy supported by CRTF".format(self.coordsys))

    def check_ds9(self):
        """
        Checks for DS9 compatibility.
        """
        if self.region_type not in regions_attributes:
            raise ValueError("'{0}' is not a valid region type in this package"
                             "supported by DS9".format(self.region_type))

        if self.coordsys not in valid_coordsys['DS9']:
            raise ValueError("'{0}' is not a valid coordinate reference frame "
                             "in astropy supported by DS9".format(self.coordsys))

    def _validate(self):
        """
        Checks whether all the attributes of this object is valid.
        """
        if self.region_type not in regions_attributes:
            raise ValueError("'{0}' is not a valid region type in this package"
                             .format(self.region_type))

        if self.coordsys not in valid_coordsys['DS9'] + valid_coordsys['CRTF']:
            raise ValueError("'{0}' is not a valid coordinate reference frame "
                             "in astropy".format(self.coordsys))


def to_shape_list(region_list, coordinate_system='fk5'):
    """
    Converts a list of regions into a `regions.ShapeList` object.

    Parameters
    ----------
    region_list: python list
        Lists of `regions.Region` objects
    format_type: str ('DS9' or 'CRTF')
        The format type of the Shape object. Default is 'DS9'.
    coordinate_system: str
        The astropy coordinate system frame in which all the coordinates present
        in the `region_list` will be converted. Default is 'fk5'.

    Returns
    -------
    shape_list: `regions.ShapeList` object
        list of `regions.Shape` objects.
    """

    shape_list = ShapeList()

    for region in region_list:

        coord = []
        if isinstance(region, SkyRegion):
            reg_type = region.__class__.__name__[:-9].lower()
        else:
            reg_type = region.__class__.__name__[:-11].lower()

        for val in regions_attributes[reg_type]:
            coord.append(getattr(region, val))

        if reg_type == 'polygon':
            coord = [x for x in region.vertices]

        if coordinate_system:
            coordsys = coordinate_system
        else:
            if isinstance(region, SkyRegion):
                coordsys = coord[0].name
            else:
                coordsys = 'image'

        frame = coordinates.frame_transform_graph.lookup_name(coordsys)

        new_coord = []
        for val in coord:
            if isinstance(val, Angle) or isinstance(val, u.Quantity) or isinstance(val, numbers.Number):
                new_coord.append(val)
            elif isinstance(val, PixCoord):
                new_coord.append(u.Quantity(val.x, u.dimensionless_unscaled))
                new_coord.append(u.Quantity(val.y, u.dimensionless_unscaled))
            else:
                new_coord.append(Angle(val.transform_to(frame).spherical.lon))
                new_coord.append(Angle(val.transform_to(frame).spherical.lat))

        meta = dict(region.meta)
        meta.update(region.visual)

        if reg_type == 'text':
            meta['text'] = meta.get('text', meta.pop('label', ''))

        include = region.meta.pop('include', True)

        shape_list.append(Shape(coordsys, reg_type, new_coord, meta, False,
                                include))

    return shape_list


def to_ds9_meta(shape_meta):
    """
    Makes the meta data DS9 compatible by filtering and mapping the valid keys

    Parameters
    ----------
    shape_meta: dict
        meta attribute of a `regions.Shape` object

    Returns
    -------
    meta : dict
        DS9 compatible meta dictionary
    """

    # meta keys allowed in DS9.
    valid_keys = ['symbol', 'include', 'tag', 'line', 'comment',
                  'name', 'select', 'highlite', 'fixed', 'label', 'text',
                  'edit', 'move', 'rotate', 'delete', 'source', 'background']

    # visual keys allowed in DS9
    valid_keys += ['color', 'dash', 'linewidth', 'font', 'dashlist',
                   'fill', 'textangle', 'symsize']

    # mapped to actual names in DS9
    key_mappings = {'symbol': 'point', 'linewidth': 'width', 'label': 'text'}

    meta = _to_io_meta(shape_meta, valid_keys, key_mappings)

    if 'font' in meta:
        meta['font'] += " {0} {1} {2}".format(shape_meta.get('fontsize', 12),
                                              shape_meta.get('fontstyle', 'normal'),
                                              shape_meta.get('fontweight', 'roman'))

    return meta


def to_crtf_meta(shape_meta):
    """
    Makes the meta data CRTF compatible by filtering and mapping the valid keys

    Parameters
    ----------
    shape_meta: dict
        meta attribute of a `regions.Region` object

    Returns
    -------
    meta : dict
        CRTF compatible meta dictionary

    """
    # please refer : https://casaguides.nrao.edu/index.php/CASA_Region_Format

    # meta keys allowed in CRTF
    valid_keys = ['label', 'include', 'frame', 'range', 'veltype',
                  'restfreq', 'coord', 'type', 'text', 'corr']

    # visual keys allowed in CRTF
    valid_keys += ['color', 'width', 'font', 'symthick', 'symsize', 'fontsize',
                   'fontstyle', 'usetex', 'labelpos', 'labeloff', 'linewidth',
                   'linestyle', 'symbol']

    key_mappings = {}

    return _to_io_meta(shape_meta, valid_keys, key_mappings)


def _to_io_meta(shape_meta, valid_keys, key_mappings):
    """
    This is used to make meta data compatible with a specific io
    by filtering and mapping to it's valid keys

    Parameters
    ----------
    shape_meta: dict
        meta attribute of a `regions.Region` object
    valid_keys : python list
        Contains all the valid keys of a particular file format.
    key_mappings : python dict
        Maps to the actual name of the key in the format.

    Returns
    -------
    meta : dict
        io compatible meta dictionary according to valid_keys and key_mappings
    """

    meta = dict()

    for key in shape_meta:
        if key in valid_keys:
            meta[key_mappings.get(key, key)] = shape_meta[key]

    return meta
